# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pandas as pd
import numpy as np
from tqdm import tqdm
import concurrent.futures

from ..functions import *

len_col_lst = [len_a_col, len_b_col, len_c_col]
ang_col_lst = [ang_a_col, ang_b_col, ang_g_col]
len_letter_lst = ["a", "b", "c"]
ang_letter_lst = ["alpha", "beta", "gamma"]


def get_path_len_ang_df(coord_path):

    df = pd.DataFrame()

    cif_dict = load_cif_dict(coord_path)

    df.at[0, rcsb_path_col] = coord_path

    if search_cif_dict(cif_dict, "_exptl.method") == "X-RAY DIFFRACTION":

        df.at[0, space_col] = search_cif_dict(
            cif_dict, "_symmetry.space_group_name_H-M"
        )

        for col, letter in zip(len_col_lst, len_letter_lst):

            df.at[0, col] = search_cif_dict(cif_dict, f"_cell.length_{letter}")

        for col, letter in zip(ang_col_lst, ang_letter_lst):

            df.at[0, col] = search_cif_dict(cif_dict, f"_cell.angle_{letter}")

    else:
        df.at[0, space_col] = "None"
        for col, letter in zip(len_col_lst, len_letter_lst):
            df.at[0, col] = 999.00
        for col, letter in zip(ang_col_lst, ang_letter_lst):
            df.at[0, col] = 999.00

    return df


def add_cf(df, min_simi=0.05):

    val_dict = dict()
    prev_dict = dict()

    col_lst = len_col_lst + ang_col_lst

    for index in tqdm(
        list(df.index.values), desc="Annotating crystal forms", position=0, leave=True
    ):

        curr_path = df.at[index, rcsb_path_col]
        curr_space = df.at[index, space_col]

        if curr_space == "None":
            df.at[index, cf_col] = "None"
        else:
            val_dict[curr_path] = {
                len_a_col: df.at[index, len_a_col],
                len_b_col: df.at[index, len_b_col],
                len_c_col: df.at[index, len_c_col],
                ang_a_col: df.at[index, ang_a_col],
                ang_b_col: df.at[index, ang_b_col],
                ang_g_col: df.at[index, ang_g_col],
            }

            if curr_space not in list(prev_dict.keys()):
                id = 1
                prev_dict[curr_space] = {id: type_lst(curr_path)}

            else:

                new_form = True

                for id in list(prev_dict[curr_space].keys()):

                    prev_path_lst = prev_dict[curr_space][id]

                    diff_form = False

                    for col in col_lst:

                        diff_lst = list()

                        for prev_path in prev_path_lst:

                            curr_val = float(val_dict[curr_path][col])
                            prev_val = float(val_dict[prev_path][col])

                            diff_lst.append(abs(prev_val - curr_val) / prev_val)

                        mean_diff = np.mean(diff_lst)

                        if mean_diff > min_simi:
                            diff_form = True
                            break

                    if not diff_form:

                        prev_dict[curr_space][id].append(curr_path)

                        new_form = False

                        break

                if new_form:

                    id = max(list(prev_dict[curr_space].keys())) + 1

                    prev_dict[curr_space][id] = type_lst(curr_path)

            df.at[index, cf_col] = f"{curr_space} (CF{id})"

    return df


def annot_cf(coord_paths, cf_table_path=None, min_simi=0.05, data=None, num_cpu=1):

    coord_path_lst = type_lst(coord_paths)

    df = pd.DataFrame()

    if num_cpu == 1:
        for coord_path in tqdm(
            coord_path_lst, desc="Getting crystal information", position=0, leave=True
        ):

            df = pd.concat([df, get_path_len_ang_df(coord_path)], sort=False)
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(get_path_len_ang_df, coord_path)
                for coord_path in coord_path_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Getting crystal information",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                df = pd.concat([df, job.result()], sort=False)

    df = df.reset_index(drop=True)

    df = add_cf(df, min_simi=min_simi)

    if data is not None:
        df_col_lst = list(data.columns)

        for col in [cf_col]:
            if col in df_col_lst:
                del data[col]

        df = merge_tables(df, data)

    print("Annotated crystal forms!")

    if cf_table_path is not None:
        save_table(cf_table_path, df)
    else:
        return df


# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import cairosvg

from rdkit import Chem
from rdkit.Chem import (
    Draw,
    rdFMCS,
    AllChem,
    PyMol,
)

from tqdm import tqdm

from .lst import type_lst
from .color import get_lst_colors, get_rgb, change_hex_alpha
from .download import download_file
from .path import path_exists, get_lig_path, append_file_path, lig_str
from .url import lig_expo_url


def download_lig(lig, lig_dir=None, check=True):

    lig_path = get_lig_path(lig, dir_path=lig_dir)

    download_file(
        f"{lig_expo_url}{lig[0]}/{lig}/{lig}_model.sdf", lig_path, check=check
    )

    return lig_path


def load_lig(lig, lig_dir=None, tries=3):

    lig_path = get_lig_path(lig, dir_path=lig_dir)

    if not path_exists(lig_path):
        lig_path = download_lig(lig, lig_dir=lig_dir)

    try_load = True
    count = 0
    while try_load:
        try:
            mol = Chem.MolFromMolFile(lig_path)
            try_load = False
        except:
            lig_path = download_lig(lig, lig_dir=lig_dir, check=False)
            count += 1
            if tries == count:
                try_load = False

    return mol


def get_lig_smiles(lig, lig_dir=None):

    mol = lig
    if type(lig) == str:
        smiles = mol
        if len(lig) <= 3:
            mol = load_lig(lig, lig_dir=lig_dir)
    if type(mol) != str:
        smiles = Chem.MolToSmiles(mol)

    return smiles


def get_lig_mol(lig, lig_dir=None, stereo=True):

    mol = lig
    if type(lig) == str:
        smiles = mol
        if len(lig) <= 3:
            smiles = get_lig_smiles(lig, lig_dir=lig_dir)
        mol = Chem.MolFromSmiles(smiles)

    if not stereo:
        AllChem.Compute2DCoords(mol)

    return mol


def get_lig_simi(lig_1, lig_2, lig_dir=None):

    mol_1 = Chem.RDKFingerprint(get_lig_mol(lig_1, lig_dir=lig_dir))
    mol_2 = Chem.RDKFingerprint(get_lig_mol(lig_2, lig_dir=lig_dir))

    return Chem.DataStructs.FingerprintSimilarity(mol_1, mol_2)


def is_lig_match(lig, matches, lig_dir=None):

    mol = get_lig_mol(lig, lig_dir=lig_dir)

    match_lst = type_lst(matches)

    if True in [
        mol.HasSubstructMatch(get_lig_mol(x, lig_dir=lig_dir)) for x in match_lst
    ]:
        return True
    else:
        return False


def get_lig_mcs(
    lig_lst,
    lig_dir=None,
    min_simi=0.9,
    compare_bonds=False,
    compare_exact=False,
    match_valence=False,
    match_rings=False,
    complete_rings=False,
):

    mol_lst = list()
    for lig in tqdm(
        lig_lst,
        desc="Loading ligands",
        position=0,
        leave=True,
    ):

        mol_lst.append(get_lig_mol(lig, lig_dir=lig_dir))

    if compare_bonds:
        if compare_exact:
            bond_compare = rdFMCS.BondCompare.CompareOrderExact
        else:
            bond_compare = rdFMCS.BondCompare.CompareOrder
    else:
        bond_compare = rdFMCS.BondCompare.CompareAny

    return Chem.MolFromSmarts(
        rdFMCS.FindMCS(
            mol_lst,
            threshold=min_simi,
            bondCompare=bond_compare,
            matchValences=match_valence,
            ringMatchesRingOnly=match_rings,
            completeRingsOnly=complete_rings,
        ).smartsString
    )


def get_lig_query_match(lig, query, lig_dir=None):

    mol = get_lig_mol(lig, lig_dir=lig_dir)
    query = get_lig_mol(query, lig_dir=lig_dir)

    atom_lst = list(mol.GetSubstructMatch(query))
    bond_lst = list()

    if len(atom_lst) > 0:

        bond_lst += [
            mol.GetBondBetweenAtoms(
                atom_lst[x.GetBeginAtomIdx()], atom_lst[x.GetEndAtomIdx()]
            ).GetIdx()
            for x in query.GetBonds()
        ]

    return atom_lst, bond_lst


def draw_lig_plot(
    ligs,
    plot_path,
    lig_dir=None,
    lig_labels=None,
    highlight_querys=None,
    highlight_alpha=0.75,
    color_palette=None,
    n_cols=None,
    mol_pad=0.1,
    plot_height=3,
    plot_width=3,
    font_size=2,
):

    lig_lst = type_lst(ligs)
    highlight_query_lst = type_lst(highlight_querys)

    if type(lig_labels) == dict:
        label_lst = [lig_labels[x] for x in lig_lst]
    else:
        label_lst = type_lst(lig_labels)
    if len(label_lst) != len(lig_lst):
        label_lst = None

    mol_lst = [get_lig_mol(x, lig_dir=lig_dir) for x in lig_lst]

    final_atom_lst = list()
    final_bond_lst = list()
    final_atom_color_lst = list()
    final_bond_color_lst = list()

    if highlight_querys is not None:
        if type(color_palette) == dict:
            color_dict = color_palette
            for key, val in color_dict.items():
                color_dict[key] = get_rgb(change_hex_alpha(val, highlight_alpha))
        else:
            color_dict = get_lst_colors(
                highlight_query_lst,
                palette=color_palette,
                return_rgb=True,
                return_dict=True,
                alpha=highlight_alpha,
            )

    for mol in mol_lst:

        mol_atom_lst = list()
        mol_bond_lst = list()
        mol_atom_color_dict = dict()
        mol_bond_color_dict = dict()

        if highlight_querys is not None:
            for i, highlight_query in enumerate(highlight_query_lst):

                atom_lst, bond_lst = get_lig_query_match(mol, highlight_query)

                if len(atom_lst) > 0:

                    mol_atom_lst += atom_lst
                    mol_bond_lst += bond_lst

                    for atom in atom_lst:
                        mol_atom_color_dict[atom] = color_dict[highlight_query]

                    for bond in bond_lst:
                        mol_bond_color_dict[bond] = color_dict[highlight_query]

        final_atom_lst.append(mol_atom_lst)
        final_bond_lst.append(mol_bond_lst)
        final_atom_color_lst.append(mol_atom_color_dict)
        final_bond_color_lst.append(mol_bond_color_dict)

    total_mols = len(mol_lst)
    if n_cols is None:
        n_cols = total_mols
    plot_size = (plot_width * 72, plot_height * 72)
    n_rows = total_mols // n_cols
    if total_mols % n_cols:
        n_rows += 1

    total_size = (n_cols * plot_size[0], n_rows * plot_size[1])
    draw = Draw.rdMolDraw2D.MolDraw2DSVG(
        total_size[0], total_size[1], plot_size[0], plot_size[1]
    )

    if n_cols is None:
        n_cols = len(mol_lst)

    draw.drawOptions().legendFontSize = font_size
    draw.drawOptions().padding = mol_pad
    draw.SetFontSize(1)
    draw.SetLineWidth(1)
    draw.DrawMolecules(
        mol_lst,
        legends=label_lst,
        highlightAtoms=final_atom_lst,
        highlightBonds=final_bond_lst,
        highlightAtomColors=final_atom_color_lst,
        highlightBondColors=final_bond_color_lst,
    )
    draw.FinishDrawing()
    img = draw.GetDrawingText()
    append_file_path(plot_path)
    cairosvg.svg2pdf(bytestring=img.encode("utf-8"), write_to=plot_path)


def show_pymol_lig(
    lig,
    lig_dir=None,
):

    lig_path = download_lig(lig, lig_dir=lig_dir)

    mv = PyMol.MolViewer()

    mv.LoadFile(lig_path, lig_str)

    mv.Zoom(lig_str)



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import gzip
import numpy as np

from Bio.PDB import (
    MMCIFParser,
    PDBParser,
    MMCIF2Dict,
    MMCIFIO,
    PDBIO,
    Selection,
    NeighborSearch,
    Superimposer,
    HSExposure,
    calc_angle,
)

from Bio.Data.SCOPData import protein_letters_3to1 as aa3to1

import math

from .seq import pair_seq_aln
from .table import (
    replace_str,
    extract_int,
    extract_str,
    make_dict,
    is_str,
    is_int,
    fix_val,
)
from .lst import (
    lst_nums,
    res_to_lst,
    str_to_lst,
    type_lst,
    sort_lst,
    lst_unique,
    lst_inter,
)


aa_lst = [
    "CYS",
    "ASP",
    "SER",
    "GLN",
    "LYS",
    "ILE",
    "PRO",
    "THR",
    "PHE",
    "ASN",
    "GLY",
    "HIS",
    "LEU",
    "ARG",
    "TRP",
    "ALA",
    "VAL",
    "GLU",
    "TYR",
    "MET",
]

bb_hb_atomid_dict = dict()
for aa in aa_lst:
    bb_hb_atomid_dict[aa] = ["O", "N"]

h_bb_hb_atomid_dict = dict()
for aa in aa_lst:
    h_bb_hb_atomid_dict[aa] = ["O", "H"]

sc_hb_atomid_dict = {
    "ASP": ["OD1"],
    "GLU": ["OE1"],
    "ASN": ["OD1"],
    "SER": ["OG"],
    "THR": ["OG1"],
    "TYR": ["OH"],
    "LYS": ["NZ"],
    "ARG": ["NE", "NH1", "NH2"],
    "GLU": ["OE1", "OE2"],
    "GLN": ["OE1", "NE2"],
    "ASN": ["ND2"],
    "HIS": ["ND1", "NE2"],
}

sc_hb_atomid_lst = list(sc_hb_atomid_dict.keys())


h_sc_hb_atomid_dict = {
    "ASP": ["OD1"],
    "GLU": ["OE1"],
    "ASN": ["OD1", "HD21", "HD22"],
    "SER": ["HG"],
    "THR": ["HG1"],
    "TYR": ["HH"],
    "LYS": ["HZ1", "HZ2", "HZ3"],
    "ARG": ["HE", "HH11", "HH12" "HH21", "HH22"],
    "GLU": ["OE1", "OE2"],
    "GLN": ["OE1", "HE21", "HE22"],
    "HIS": ["ND1", "NE2", "HD1", "HE2"],
}

hb_name = "H-Bond"
wmhb_name = "WMH-Bond"
no_hb_name = "No Bond"

disorder_dict = {2: True, 1: True, 0: False}


def get_pdb_id(pdb_code, chainid):

    return f"{pdb_code}{chainid}"


def get_pdb_code(pdb_id):

    if len(pdb_id) == 5:
        pdb_id = pdb_id[0:4]

    return pdb_id


def get_pdb_chainid(pdb_id):

    if len(pdb_id) == 5:
        pdb_id = pdb_id[4:5]

    return pdb_id


def build_pdb_code_lst(pdb_id_lst):

    pdb_code_lst = list()
    for pdb_id in pdb_id_lst:
        pdb_code = get_pdb_code(pdb_id)
        if pdb_code not in pdb_code_lst:
            pdb_code_lst.append(pdb_code)

    return pdb_code_lst


def build_pdb_dict(pdb_id_lst):

    pdb_dict = {}

    for pdb_id in pdb_id_lst:

        pdb_code = get_pdb_code(pdb_id)
        chainid = get_pdb_chainid(pdb_id)

        if pdb_code not in pdb_dict.keys():
            pdb_dict[pdb_code] = str_to_lst(chainid)
        else:
            pdb_dict[pdb_code].append(chainid)

    return pdb_dict


def load_coord(path):

    if ".cif" in path:
        parser = MMCIFParser(QUIET=True)

    elif ".pdb" in path:
        parser = PDBParser(QUIET=True)

    if ".gz" in path:
        path = gzip.open(path, "rt")

    return parser.get_structure("PARSER", path)


def load_cif_dict(path):

    if ".gz" in path:
        path = gzip.open(path, "rt")

    return MMCIF2Dict.MMCIF2Dict(path)


def search_cif_dict(cif_dict, key):

    return replace_str(str(cif_dict[key][0]), ["?", "-1"], replace=None)


def save_coord(path, structure, sele=None):

    if ".cif" in path:

        io = MMCIFIO()

    elif ".pdb" in path:

        io = PDBIO()

    io.set_structure(structure)

    if sele is not None:
        io.save(path, sele)
    else:
        io.save(path)


def resname_to_letter(resname):

    return aa3to1.get(resname, "X")


def get_resname(residue, letter=False):

    resname = str(residue.get_resname())

    resname = resname.strip()

    if letter:
        resname = resname_to_letter(resname)

    return resname


def unfold_structure(structure, level="A"):

    return Selection.unfold_entities(structure, level)


def get_models(structure):

    return unfold_structure(structure, level="M")


def get_chains(structure):

    return unfold_structure(structure, level="C")


def get_residues(structure):

    return unfold_structure(structure, level="R")


def get_atoms(structure):

    return unfold_structure(structure, level="A")


def get_parent(structure):

    return structure.get_parent()


def get_modelid(model):

    return int(model.id)


def get_chainid(chain):

    return str(chain.id)


def get_resid(residue):

    return residue.id


def get_resmodelid(residue):

    return int(get_modelid(get_parent(get_parent(residue))))


def get_reschainid(residue):

    return str(get_chainid(get_parent(residue)))


def get_resnum(residue):

    return int(get_resid(residue)[1])


def get_resid_str(residue):

    resid = get_resid(residue)

    return f"{resid[1]}{resid[2]}".replace(" ", "")


def resid_to_tuple(resid):

    if type(resid) != tuple:
        resnum = extract_int(resid)
        insert = extract_str(resid)
        if insert == "":
            insert = " "
        resid = tuple([" ", resnum, insert])

    return resid


def get_atomid(atom):

    return str(atom.id)


def is_aa(residue):

    return get_resid(residue)[0] == " "


def is_wat(residue):

    return get_resid(residue)[0] == "W"


def is_het(residue):

    return not is_aa(residue) and not is_wat(residue)


def is_disordered(residue):

    return disorder_dict[residue.is_disordered()]


def has_modelid(structure, modelid):

    if modelid is None:
        modelid = 0

    return structure.has_id(fix_val(modelid, return_int=True))


def has_chainid(structure, chainid, modelid=None):

    if modelid is None:
        modelid = 0

    chainid_status = False
    if has_modelid(structure, modelid):
        chainid_status = structure[fix_val(modelid, return_int=True)].has_id(chainid)

    return chainid_status


def has_resid(structure, chainid, resid, modelid=None):

    if modelid is None:
        modelid = 0

    resid_status = False
    if has_chainid(structure, chainid, modelid=modelid):
        resid_status = structure[fix_val(modelid, return_int=True)][chainid].has_id(
            resid
        )
    return resid_status


def has_atomid(structure, chainid, resid, atomid, modelid=None):

    if modelid is None:
        modelid = 0

    atomid_status = False
    if has_resid(structure, chainid, resid, modelid=modelid):
        atomid_status = structure[fix_val(modelid, return_int=True)][chainid][
            resid
        ].has_id(atomid)
    return atomid_status


def has_altloc(structure, chainid, resid, atomid, altloc, modelid=None):

    if modelid is None:
        modelid = 0

    altloc_status = False

    if has_atomid(structure, chainid, resid, atomid, modelid=modelid):
        atom = structure[fix_val(modelid, return_int=True)][chainid][resid][atomid]
        if is_disordered(atom):
            altloc_status = atom.disordered_has_id(altloc)

    return altloc_status


def get_neighbors(structure):

    return NeighborSearch(get_atoms(structure))


def get_chain_cont(neighbors, chain, max_dist=5, level="A", remove_disordered=False):

    atom_lst = list()

    for residue in chain:

        atom_lst += get_residue_cont(
            neighbors,
            residue,
            max_dist=max_dist,
            level=level,
            remove_disordered=remove_disordered,
        )

    return atom_lst


def get_residue_cont(
    neighbors, residue, max_dist=5, level="A", remove_disordered=False
):

    atom_lst = list()

    for atom in residue:

        atom_lst += get_atom_cont(
            neighbors,
            atom,
            max_dist=max_dist,
            level=level,
            remove_disordered=remove_disordered,
        )

    return atom_lst


def get_atom_cont(neighbors, atom, max_dist=5, level="A", remove_disordered=False):

    atom_lst = neighbors.search(atom.get_coord(), max_dist, level=level)

    if remove_disordered:
        return [x for x in atom_lst if not is_disordered(x)]
    else:
        return atom_lst


def get_seq_lst(structure):

    aa_info = lambda r: (r.id[1], resname_to_letter(get_resname(r)))
    return [aa_info(r) for r in get_residues(structure) if is_aa(r)]


def join_seq_lst(seq_lst):

    return "".join([i[1] for i in seq_lst])


def map_seqs(
    ref_seq_lst,
    mob_seq_lst,
):

    ref_seq = join_seq_lst(ref_seq_lst)
    mob_seq = join_seq_lst(mob_seq_lst)

    aln = pair_seq_aln(ref_seq, mob_seq)

    best_aln = aln[0]

    aln_ref = best_aln[0]
    aln_mob = best_aln[1]

    map_dict = {}
    aa_i_ref, aa_i_mob = 0, 0
    for aa_aln_ref, aa_aln_mob in zip(aln_ref, aln_mob):
        if aa_aln_ref == "-":
            if aa_aln_mob != "-":
                aa_i_mob += 1
        elif aa_aln_mob == "-":
            if aa_aln_ref != "-":
                aa_i_ref += 1
        else:
            assert ref_seq_lst[aa_i_ref][1] == aa_aln_ref
            assert mob_seq_lst[aa_i_mob][1] == aa_aln_mob
            map_dict[ref_seq_lst[aa_i_ref][0]] = mob_seq_lst[aa_i_mob][0]
            aa_i_ref += 1
            aa_i_mob += 1

    return map_dict


def build_map_dict(
    ref_chain,
    mob_chain,
    ref_seq_lst=None,
    mob_seq_lst=None,
    pair_aln=False,
):

    if pair_aln:
        if ref_seq_lst == None:
            ref_seq_lst = get_seq_lst(ref_chain)

        if mob_seq_lst == None:
            mob_seq_lst = get_seq_lst(mob_chain)

        map_dict = map_seqs(
            ref_seq_lst,
            mob_seq_lst,
        )
    else:
        ref_resid_lst = [get_resnum(residue) for residue in ref_chain if is_aa(residue)]
        mob_resid_lst = [get_resnum(residue) for residue in mob_chain if is_aa(residue)]

        resid_lst = [x for x in ref_resid_lst if x in mob_resid_lst]

        map_dict = make_dict(resid_lst, resid_lst)

    return map_dict


def remap_dict(map_dict, resids):

    resid_lst = res_to_lst(resids)

    remap_dict = dict()

    for ref_resid, mob_resid in map_dict.items():
        if ref_resid in resid_lst:
            remap_dict[ref_resid] = mob_resid

    return remap_dict


def get_sup_atoms(ref_chain, mob_chain, map_dict, sup_resids=None):

    if sup_resids is not None:
        map_dict = remap_dict(map_dict, sup_resids)

    ref_ca_lst = list()
    mob_ca_lst = list()

    for ref_resid in map_dict:
        if ref_chain[ref_resid].has_id("CA") and mob_chain[map_dict[ref_resid]].has_id(
            "CA"
        ):
            ref_ca_lst.append(ref_chain[ref_resid]["CA"])
            mob_ca_lst.append(mob_chain[map_dict[ref_resid]]["CA"])

    return (ref_ca_lst, mob_ca_lst)


def sup_coord(mob_structure, ref_ca_lst, mob_ca_lst):

    sup = Superimposer()
    sup.set_atoms(ref_ca_lst, mob_ca_lst)
    sup.apply(mob_structure.get_atoms())

    return mob_structure


def sup_with_map(ref_chain, mob_chain, map_dict, sup_resids=None):

    sup_atoms = get_sup_atoms(ref_chain, mob_chain, map_dict, sup_resids=sup_resids)

    sup_chain = sup_coord(mob_chain, sup_atoms[0], sup_atoms[1])

    return (ref_chain, sup_chain)


def sup_without_map(
    ref_chain,
    mob_chain,
    ref_seq_lst=None,
    mob_seq_lst=None,
    pair_aln=False,
    sup_resids=None,
):

    map_dict = build_map_dict(
        ref_chain,
        mob_chain,
        ref_seq_lst=ref_seq_lst,
        mob_seq_lst=mob_seq_lst,
        pair_aln=pair_aln,
    )

    ref_chain, sup_chain = sup_with_map(
        ref_chain, mob_chain, map_dict, sup_resids=sup_resids
    )

    return (ref_chain, sup_chain, map_dict)


def calc_norm_dist(vect_1, vect_2, dist_dec=2):

    return round(np.linalg.norm(vect_1 - vect_2), dist_dec)


def prep_resid(structure, chainid, resid, modelid=None):

    if modelid is None:
        modelid = 0

    if is_str(resid):
        for residue in structure[fix_val(modelid, return_int=True)][chainid]:
            if is_het(residue):
                if resid == get_resname(residue):
                    resid = get_resid(residue)
    elif is_int(resid):
        resid = resid_to_tuple(resid)

    return resid


def select_atom_altloc(atom, altloc):

    if altloc is not None:
        atom.disordered_select(altloc)

    return atom


def get_atom_vect(
    structure,
    chainid,
    resid,
    atomid,
    modelid=None,
    altloc=None,
    coord=False,
):

    if modelid is None:
        modelid = 0

    atom_vect = 999.00

    resid = prep_resid(structure, chainid, resid, modelid=modelid)

    if has_atomid(structure, chainid, resid, atomid, modelid=modelid):
        atom = structure[fix_val(modelid, return_int=True)][chainid][resid][atomid]
        if altloc is not None:
            if has_altloc(structure, chainid, resid, atomid, altloc, modelid=modelid):
                atom = select_atom_altloc(atom, altloc)
        if coord:
            atom_vect = atom.get_coord()
        else:
            atom_vect = atom.get_vector()

    return atom_vect


def calc_rmsd(
    ref_chain,
    sup_chain,
    map_dict=None,
    rmsd_resids=None,
    rmsd_atomids="CA",
    pair_aln=False,
):

    if map_dict is None:
        map_dict = build_map_dict(ref_chain, sup_chain, pair_aln=pair_aln)

    map_resid_lst = list(map_dict.keys())

    if rmsd_resids is None:
        resid_lst = map_resid_lst
    else:
        resid_lst = res_to_lst(rmsd_resids)

    atomid_lst = type_lst(rmsd_atomids)

    ref_vect_lst = list()
    sup_vect_lst = list()

    for resid in resid_lst:
        if resid in map_resid_lst:
            ref_resid = resid_to_tuple(resid)
            sup_resid = resid_to_tuple(map_dict[resid])
            for atomid in atomid_lst:
                ref_vect = 999.00
                if ref_chain[ref_resid].has_id(atomid):
                    ref_vect = ref_chain[ref_resid][atomid].get_coord()
                if type(ref_vect) != float:
                    sup_vect = 999.00
                    if sup_chain[sup_resid].has_id(atomid):
                        sup_vect = sup_chain[sup_resid][atomid].get_coord()
                    if type(sup_vect) != float:
                        ref_vect_lst.append(list(ref_vect))
                        sup_vect_lst.append(list(sup_vect))

    tot_vects = len(ref_vect_lst)

    if tot_vects == 0:
        rmsd_val = 999.00
    else:
        ref_vects = np.array(ref_vect_lst)
        sup_vects = np.array(sup_vect_lst)
        diff_vect = ref_vects - sup_vects
        rmsd_val = float(np.sqrt((diff_vect ** 2).sum() / tot_vects))

    return rmsd_val


def build_add_resid_lst(
    coord_path,
    modelid,
    chainid,
    resid_lst,
    dict_resid_lst,
    max_ca_dist=4,
    ext_mult=1,
):

    min_resid = resid_lst[0]
    max_resid = resid_lst[-1]

    search_resid_lst = resid_lst.copy()

    search_ext = len(resid_lst * ext_mult)

    search_lst = lst_nums(1, search_ext)

    for i in search_lst:
        search_resid_lst.append(min_resid - i)
        search_resid_lst.append(max_resid + i)

    search_resid_lst = sort_lst(search_resid_lst)

    search_resid_dict = dict()

    for search_resid in search_resid_lst:
        search_resid_dict[search_resid] = [
            x for x in dict_resid_lst if extract_int(x) == extract_int(search_resid)
        ]

    loaded = False
    add_resid_lst = list()
    for resid in resid_lst:
        curr_resid_lst = search_resid_dict[resid]
        if len(curr_resid_lst) == 0:
            prev_resid = None
            next_resid = None
            for i in search_lst:
                prev_resid_lst = search_resid_dict[resid - i]
                next_resid_lst = search_resid_dict[resid + i]
                if prev_resid is None:
                    if len(prev_resid_lst) > 0:
                        prev_resid = prev_resid_lst[-1]
                        curr_resid = extract_int(prev_resid) + i
                if next_resid is None:
                    if len(next_resid_lst) > 0:
                        next_resid = next_resid_lst[0]
                        curr_resid = extract_int(next_resid) - i
                if prev_resid is not None and next_resid is not None:
                    if not loaded:
                        structure = load_coord(coord_path)
                        loaded = True

                    atom_dist = calc_atom_dist(
                        structure=structure,
                        chainid_1=chainid,
                        resid_1=prev_resid,
                        chainid_2=chainid,
                        resid_2=next_resid,
                        atomid_1="CA",
                        atomid_2="CA",
                        modelid_1=modelid,
                        modelid_2=modelid,
                    )
                    if atom_dist > max_ca_dist:
                        curr_resid_lst = type_lst(str(curr_resid))
                    break

        for curr_resid in curr_resid_lst:
            if curr_resid not in add_resid_lst:
                add_resid_lst.append(curr_resid)

    return add_resid_lst


def get_hb_atomid_lst(
    structure, chainid, resid, modelid=None, hb_sc=True, hb_bb=True, use_h=False
):

    if modelid is None:
        modelid = 0

    atomid_lst = list()

    used_bb_hb_atomid_dict = bb_hb_atomid_dict.copy()
    used_sc_hb_atomid_dict = sc_hb_atomid_dict.copy()
    if use_h:
        used_bb_hb_atomid_dict = h_bb_hb_atomid_dict.copy()
        used_sc_hb_atomid_dict = h_sc_hb_atomid_dict.copy()

    if has_resid(structure, chainid, resid, modelid=modelid):

        residue = structure[fix_val(modelid, return_int=True)][chainid][resid]

        resname = get_resname(residue)

        if is_aa(residue):

            atomid_lst = list()

            if hb_bb is not False:
                if resname in aa_lst:
                    atomid_lst += used_bb_hb_atomid_dict[resname]
                    if type(hb_bb) == str:
                        atomid_lst = [x for x in atomid_lst if hb_bb in x]
            if hb_sc is not False:
                if resname in sc_hb_atomid_lst:
                    atomid_lst += used_sc_hb_atomid_dict[resname]
                    if type(hb_sc) == str:
                        atomid_lst = [x for x in atomid_lst if hb_sc in x]
        else:
            for atom in residue:
                atomid = get_atomid(atom)
                if atomid[0] == "O" or atomid[0] == "N" or atomid[0] == "H":
                    atomid_lst.append(atomid)

    if len(atomid_lst) == 0:
        atomid_lst = None

    return atomid_lst


def pred_wmhb_dist(max_wmhb_dist, max_wmhb_angle):

    return np.sqrt(
        (max_wmhb_dist ** 2)
        + (max_wmhb_dist ** 2)
        - (2 * max_wmhb_dist * max_wmhb_dist * math.cos(math.radians(max_wmhb_angle)))
    )


def get_adj_atom(
    structure, chainid, resid, atomid, altloc=None, modelid=None, adj_dist=4
):

    if modelid is None:
        modelid == 0

    adj_atom = None

    if has_resid(structure, chainid, resid, modelid=modelid):
        chain = structure[fix_val(modelid, return_int=True)][chainid]
        atom = chain[resid][atomid]

        neighbors = get_neighbors(chain)
        atom = select_atom_altloc(atom, altloc)

        atom_vect = atom.get_vector()

        adj_dist_lst = list()
        adj_atom_lst = list()

        for adj_atom in get_atom_cont(neighbors, atom, max_dist=adj_dist):
            adj_atomid = get_atomid(adj_atom)
            if adj_atomid != atomid:
                if (atomid[0] == "H") or (
                    atomid[0] != "H"
                    and (adj_atomid[0] == "C" or is_het(atom.get_parent()))
                ):
                    adj_dist_lst.append(
                        calc_norm_dist(atom_vect, adj_atom.get_vector())
                    )
                    adj_atom_lst.append(adj_atom)

        if len(adj_atom_lst) > 0:
            adj_atom = adj_atom_lst[adj_dist_lst.index(min(adj_dist_lst))]

    return adj_atom


def find_hb_status(
    atom_dist,
    structure,
    chainid_1,
    resid_1,
    atomid_1,
    chainid_2,
    resid_2,
    atomid_2,
    altloc_1=None,
    altloc_2=None,
    modelid_1=None,
    modelid_2=None,
    adj_atomid_1=None,
    adj_atomid_2=None,
    min_hb_dist=2.0,
    max_hb_dist=3.2,
    min_wmhb_dist=2.0,
    max_wmhb_dist=3.0,
    min_hb_angle=90,
    max_hb_angle=180,
    min_wmhb_angle=80,
    max_wmhb_angle=140,
    return_angle=False,
):

    if modelid_1 is None:
        modelid_1 = 0
    if modelid_2 is None:
        modelid_2 = 0

    hb_status = no_hb_name
    hb_angle_1 = 999.00
    hb_angle_2 = 999.00
    wmhb_angle = 999.00
    outlier_status = False

    min_wmhb_diag_dist = pred_wmhb_dist(min_wmhb_dist, min_wmhb_angle)
    max_wmhb_diag_dist = pred_wmhb_dist(max_wmhb_dist, max_wmhb_angle)

    if has_resid(structure, chainid_1, resid_1, modelid=modelid_1) and has_resid(
        structure, chainid_2, resid_2, modelid=modelid_2
    ):
        residue_1 = structure[fix_val(modelid_1, return_int=True)][chainid_1][resid_1]
        residue_2 = structure[fix_val(modelid_2, return_int=True)][chainid_2][resid_2]

        resname_1 = get_resname(residue_1)
        resname_2 = get_resname(residue_2)

        calc_hb = True
        if is_aa(residue_1):
            if (atomid_1 not in bb_hb_atomid_dict[resname_1]) and (
                atomid_1 not in h_bb_hb_atomid_dict[resname_1]
            ):
                calc_hb = False
                if resname_1 in sc_hb_atomid_lst:
                    if (atomid_1 in sc_hb_atomid_dict[resname_1]) or (
                        (atomid_1 in h_sc_hb_atomid_dict[resname_1])
                    ):
                        calc_hb = True

        if is_aa(residue_2):
            if (atomid_2 not in bb_hb_atomid_dict[resname_2]) and (
                atomid_2 not in h_bb_hb_atomid_dict[resname_2]
            ):
                calc_hb = False
                if resname_2 in sc_hb_atomid_lst:
                    if (atomid_2 in sc_hb_atomid_dict[resname_2]) or (
                        (atomid_2 in h_sc_hb_atomid_dict[resname_2])
                    ):
                        calc_hb = True

        if calc_hb:

            find_adj_1 = True
            if adj_atomid_1 is not None:
                if has_atomid(
                    structure, chainid_1, resid_1, adj_atomid_1, modelid=modelid_1
                ):
                    adj_atom_1 = structure[fix_val(modelid_1, return_int=True)][
                        chainid_1
                    ][resid_1][adj_atomid_1]
                    find_adj_1 = False
            if find_adj_1:
                adj_atom_1 = get_adj_atom(
                    structure,
                    chainid_1,
                    resid_1,
                    atomid_1,
                    altloc=altloc_1,
                    modelid=modelid_1,
                )

            find_adj_2 = True
            if adj_atomid_2 is not None:
                if has_atomid(
                    structure, chainid_2, resid_2, adj_atomid_2, modelid=modelid_2
                ):
                    adj_atom_2 = structure[fix_val(modelid_2, return_int=True)][
                        chainid_2
                    ][resid_2][adj_atomid_2]
                    find_adj_2 = False
            if find_adj_2:
                adj_atom_2 = get_adj_atom(
                    structure,
                    chainid_2,
                    resid_2,
                    atomid_2,
                    altloc=altloc_2,
                    modelid=modelid_2,
                )

            if adj_atom_1 is not None and adj_atom_2 is not None:

                atom_1 = select_atom_altloc(residue_1[atomid_1], altloc_1)
                atom_2 = select_atom_altloc(residue_2[atomid_2], altloc_2)

                atom_1_vect = atom_1.get_vector()
                atom_2_vect = atom_2.get_vector()
                adj_atom_1_vect = adj_atom_1.get_vector()
                adj_atom_2_vect = adj_atom_2.get_vector()

                if (min_hb_dist <= atom_dist <= max_hb_dist) and (
                    (atomid_1[0] != atomid_2[0])
                    or (resname_1 not in aa_lst or resname_2 not in aa_lst)
                ):

                    hb_angle_1 = math.degrees(
                        calc_angle(adj_atom_1_vect, atom_1_vect, atom_2_vect)
                    )

                    hb_angle_2 = math.degrees(
                        calc_angle(adj_atom_2_vect, atom_2_vect, atom_1_vect)
                    )
                    if (min_hb_angle <= hb_angle_1 <= max_hb_angle) and (
                        min_hb_angle <= hb_angle_2 <= max_hb_angle
                    ):
                        hb_status = hb_name
                    else:
                        outlier_status = True
                else:
                    neighbors_1 = get_neighbors(
                        structure[fix_val(modelid_1, return_int=True)][chainid_1]
                    )
                    neighbors_2 = get_neighbors(
                        structure[fix_val(modelid_2, return_int=True)][chainid_2]
                    )

                    atom_cont_1_lst = get_atom_cont(
                        neighbors_1, atom_1, max_dist=max_wmhb_dist
                    )
                    atom_cont_2_lst = get_atom_cont(
                        neighbors_2, atom_2, max_dist=max_wmhb_dist
                    )

                    residue_cont_1_lst = [x.get_parent() for x in atom_cont_1_lst]

                    residue_cont_2_lst = [x.get_parent() for x in atom_cont_2_lst]

                    wat_cont_lst_1 = lst_unique(
                        [x for x in residue_cont_1_lst if is_wat(x)]
                    )
                    wat_cont_lst_2 = lst_unique(
                        [x for x in residue_cont_2_lst if is_wat(x)]
                    )

                    wat_cont_lst = lst_inter(wat_cont_lst_1, wat_cont_lst_2)

                    if len(wat_cont_lst) > 0:

                        for wat_residue in wat_cont_lst:

                            wat_atom = wat_residue["O"]

                            wat_atom_vect = wat_atom.get_vector()

                            wmhb_dist_1 = calc_norm_dist(atom_1_vect, wat_atom_vect)

                            wmhb_dist_2 = calc_norm_dist(atom_2_vect, wat_atom_vect)

                            if (min_wmhb_dist < wmhb_dist_1 <= max_wmhb_dist) and (
                                min_wmhb_dist < wmhb_dist_2 <= max_wmhb_dist
                            ):

                                wmhb_angle = math.degrees(
                                    calc_angle(
                                        atom_1_vect,
                                        wat_atom_vect,
                                        atom_2_vect,
                                    )
                                )
                                if min_wmhb_angle < wmhb_angle < max_wmhb_angle:
                                    hb_angle_1 = math.degrees(
                                        calc_angle(
                                            adj_atom_1_vect,
                                            atom_1_vect,
                                            wat_atom_vect,
                                        )
                                    )
                                    hb_angle_2 = math.degrees(
                                        calc_angle(
                                            adj_atom_2_vect,
                                            atom_2_vect,
                                            wat_atom_vect,
                                        )
                                    )
                                    if (
                                        min_hb_angle <= hb_angle_1 <= max_hb_angle
                                    ) and (min_hb_angle <= hb_angle_2 <= max_hb_angle):
                                        hb_status = wmhb_name
                                        break
                                    else:
                                        outlier_status = True
                                else:
                                    outlier_status = True

                    elif min_wmhb_diag_dist < atom_dist < max_wmhb_diag_dist:
                        hb_status = wmhb_name

    if return_angle:
        return (hb_status, hb_angle_1, hb_angle_2, wmhb_angle, outlier_status)
    else:
        return hb_status


def get_altloc_lst(structure, chainid, resid, atomid, modelid=None):

    altloc_lst = [None]
    for altloc in ["A", "B", "C", "D", "E"]:
        if has_altloc(structure, chainid, resid, atomid, altloc, modelid=modelid):
            altloc_lst.append(altloc)

    return altloc_lst


def calc_atom_dist(
    structure,
    chainid_1,
    resid_1,
    chainid_2,
    resid_2,
    atomid_1=None,
    atomid_2=None,
    modelid_1=None,
    modelid_2=None,
    adj_atomid_1=None,
    adj_atomid_2=None,
    check_hb=False,
    use_h=False,
    hb_sc=True,
    hb_bb=True,
    min_hb_dist=2.0,
    max_hb_dist=3.2,
    min_wmhb_dist=2.0,
    max_wmhb_dist=3.0,
    min_hb_angle=90,
    max_hb_angle=180,
    min_wmhb_angle=80,
    max_wmhb_angle=140,
    dist_dec=2,
    return_vect=False,
):

    if modelid_1 is None:
        modelid_1 = 0
    if modelid_2 is None:
        modelid_2 = 0

    resid_1 = prep_resid(structure, chainid_1, resid_1, modelid=modelid_1)
    resid_2 = prep_resid(structure, chainid_2, resid_2, modelid=modelid_2)

    if atomid_1 is None:
        atomid_1_lst = get_hb_atomid_lst(
            structure,
            chainid_1,
            resid_1,
            modelid=modelid_1,
            hb_sc=hb_sc,
            hb_bb=hb_bb,
            use_h=use_h,
        )
    else:
        atomid_1_lst = type_lst(atomid_1)

    if atomid_2 is None:
        atomid_2_lst = get_hb_atomid_lst(
            structure,
            chainid_2,
            resid_2,
            modelid=modelid_2,
            hb_sc=hb_sc,
            hb_bb=hb_bb,
            use_h=use_h,
        )
    else:
        atomid_2_lst = type_lst(atomid_2)

    atom_dist_lst = list()
    vect_1_lst = list()
    vect_2_lst = list()

    if atomid_1_lst is not None and atomid_2_lst is not None:

        altloc_1_lst = [
            get_altloc_lst(structure, chainid_1, resid_1, x, modelid=modelid_1)
            for x in atomid_1_lst
        ]
        altloc_2_lst = [
            get_altloc_lst(structure, chainid_2, resid_2, x, modelid=modelid_2)
            for x in atomid_2_lst
        ]

        atomid_pair_lst = list()
        for i, atomid_1 in enumerate(atomid_1_lst):
            for altloc_1 in altloc_1_lst[i]:
                for j, atomid_2 in enumerate(atomid_2_lst):
                    for altloc_2 in altloc_2_lst[j]:
                        atomid_pair_lst.append((atomid_1, altloc_1, atomid_2, altloc_2))

        for atomid_pair in atomid_pair_lst:

            atomid_1 = atomid_pair[0]
            altloc_1 = atomid_pair[1]

            atomid_2 = atomid_pair[2]
            altloc_2 = atomid_pair[3]

            vect_1 = 999.00
            vect_2 = 999.00
            atom_dist = 999.00

            vect_1 = get_atom_vect(
                structure=structure,
                modelid=modelid_1,
                chainid=chainid_1,
                resid=resid_1,
                atomid=atomid_1,
                altloc=altloc_1,
            )

            if vect_1 != 999.00:

                vect_2 = get_atom_vect(
                    structure=structure,
                    modelid=modelid_2,
                    chainid=chainid_2,
                    resid=resid_2,
                    atomid=atomid_2,
                    altloc=altloc_2,
                )

                if vect_2 != 999.00:

                    atom_dist = calc_norm_dist(vect_1, vect_2, dist_dec=dist_dec)

            vect_1_lst.append(vect_1)
            vect_2_lst.append(vect_2)
            atom_dist_lst.append(atom_dist)

    atomid_pair = None
    if len(atom_dist_lst) == 0:
        vect_1 = 999.00
        vect_2 = 999.00
        atom_dist = 999.00
    else:
        atom_dist = min(atom_dist_lst)
        vect_1 = vect_1_lst[atom_dist_lst.index(atom_dist)]
        vect_2 = vect_2_lst[atom_dist_lst.index(atom_dist)]
        if atom_dist != 999.00:
            atomid_pair = atomid_pair_lst[atom_dist_lst.index(atom_dist)]

    if check_hb:

        hb_status = no_hb_name
        hb_angle_1 = 999.00
        hb_angle_2 = 999.00
        wmhb_angle = 999.00
        outlier_status = True

        if atomid_pair is not None:

            atomid_1 = atomid_pair[0]
            altloc_1 = atomid_pair[1]

            atomid_2 = atomid_pair[2]
            altloc_2 = atomid_pair[3]

            hb_result = find_hb_status(
                atom_dist,
                structure,
                chainid_1,
                resid_1,
                atomid_1,
                chainid_2,
                resid_2,
                atomid_2,
                altloc_1=altloc_1,
                altloc_2=altloc_2,
                modelid_1=modelid_1,
                modelid_2=modelid_2,
                adj_atomid_1=adj_atomid_1,
                adj_atomid_2=adj_atomid_2,
                min_hb_dist=min_hb_dist,
                max_hb_dist=max_hb_dist,
                min_wmhb_dist=min_wmhb_dist,
                max_wmhb_dist=max_wmhb_dist,
                min_hb_angle=min_hb_angle,
                max_hb_angle=max_hb_angle,
                min_wmhb_angle=min_wmhb_angle,
                max_wmhb_angle=max_wmhb_angle,
                return_angle=True,
            )

            hb_status = hb_result[0]
            hb_angle_1 = hb_result[1]
            hb_angle_2 = hb_result[2]
            wmhb_angle = hb_result[3]
            outlier_status = hb_result[4]

    result = (atom_dist,)

    if return_vect:
        result += (vect_1, vect_2)

    if check_hb:
        result += (hb_status, hb_angle_1, hb_angle_2, wmhb_angle, outlier_status)

    if len(result) == 1:
        result = result[0]

    return result


def build_hse_dict(model, radius=12):

    return HSExposure.HSExposureCA(model, radius=radius)


def get_hse_val(hse_dict, chainid, resid, index=None):

    hse_tuple = tuple([chainid, resid_to_tuple(resid)])

    hse = 999.00
    if hse_tuple in list(hse_dict.keys()):
        hse = hse_dict[hse_tuple]

    if index is not None:
        hse = hse[index]

    return hse


def get_hse_up(hse_dict, chainid, resid):

    return get_hse_val(hse_dict, chainid, resid, index=0)


def get_hse_down(hse_dict, chainid, resid):

    return get_hse_val(hse_dict, chainid, resid, index=1)



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import math
from collections import defaultdict
from Bio.PDB import calc_dihedral


from .table import extract_int, fix_val
from .lst import lst_nums
from .coord import (
    resid_to_tuple,
    get_resname,
    get_atom_vect,
    calc_norm_dist,
    has_resid,
    has_atomid,
)

from .col import (
    phi_col,
    psi_col,
    omega_col,
    chi1_col,
    altchi1_col,
    chi2_col,
    altchi2_col,
    chi3_col,
    chi4_col,
    chi5_col,
)

bb_dict = {
    phi_col: [["C", "N", "CA", "C"], [-1, 0, 0, 0]],
    psi_col: [["N", "CA", "C", "N"], [0, 0, 0, 1]],
    omega_col: [["CA", "C", "N", "CA"], [-1, -1, 0, 0]],
}

sc_dict = {
    chi1_col: {
        "ARG": ["N", "CA", "CB", "CG"],
        "ASN": ["N", "CA", "CB", "CG"],
        "ASP": ["N", "CA", "CB", "CG"],
        "CYS": ["N", "CA", "CB", "SG"],
        "GLN": ["N", "CA", "CB", "CG"],
        "GLU": ["N", "CA", "CB", "CG"],
        "HIS": ["N", "CA", "CB", "CG"],
        "ILE": ["N", "CA", "CB", "CG1"],
        "LEU": ["N", "CA", "CB", "CG"],
        "LYS": ["N", "CA", "CB", "CG"],
        "MET": ["N", "CA", "CB", "CG"],
        "PHE": ["N", "CA", "CB", "CG"],
        "PRO": ["N", "CA", "CB", "CG"],
        "SER": ["N", "CA", "CB", "OG"],
        "THR": ["N", "CA", "CB", "OG1"],
        "TRP": ["N", "CA", "CB", "CG"],
        "TYR": ["N", "CA", "CB", "CG"],
        "VAL": ["N", "CA", "CB", "CG1"],
    },
    altchi1_col: {"VAL": ["N", "CA", "CB", "CG2"]},
    chi2_col: {
        "ARG": ["CA", "CB", "CG", "CD"],
        "ASN": ["CA", "CB", "CG", "OD1"],
        "ASP": ["CA", "CB", "CG", "OD1"],
        "GLN": ["CA", "CB", "CG", "CD"],
        "GLU": ["CA", "CB", "CG", "CD"],
        "HIS": ["CA", "CB", "CG", "ND1"],
        "ILE": ["CA", "CB", "CG1", "CD1"],
        "LEU": ["CA", "CB", "CG", "CD1"],
        "LYS": ["CA", "CB", "CG", "CD"],
        "MET": ["CA", "CB", "CG", "SD"],
        "PHE": ["CA", "CB", "CG", "CD1"],
        "PRO": ["CA", "CB", "CG", "CD"],
        "TRP": ["CA", "CB", "CG", "CD1"],
        "TYR": ["CA", "CB", "CG", "CD1"],
    },
    altchi2_col: {
        "ASP": ["CA", "CB", "CG", "OD2"],
        "LEU": ["CA", "CB", "CG", "CD2"],
        "PHE": ["CA", "CB", "CG", "CD2"],
        "TYR": ["CA", "CB", "CG", "CD2"],
    },
    chi3_col: {
        "ARG": ["CB", "CG", "CD", "NE"],
        "GLN": ["CB", "CG", "CD", "OE1"],
        "GLU": ["CB", "CG", "CD", "OE1"],
        "LYS": ["CB", "CG", "CD", "CE"],
        "MET": ["CB", "CG", "SD", "CE"],
    },
    chi4_col: {
        "ARG": ["CG", "CD", "NE", "CZ"],
        "LYS": ["CG", "CD", "CE", "NZ"],
    },
    chi5_col: {"ARG": ["CD", "NE", "CZ", "NH1"]},
}


def get_rama_type(phi, psi, omega):

    phi = float(phi)
    psi = float(psi)
    omega = float(omega)

    omega = math.radians(omega)

    if phi > 180.0:
        phi = -360.0 + phi
    if psi > 180.0:
        psi = -360.0 + psi
    if phi == -180:
        phi = 180
    if psi == -180:
        psi = 180

    regions = defaultdict(list)

    regions["B"].append([-180, -100, 50, 180])
    regions["B"].append([-180, -100, -180, -100])
    regions["A"].append([-180, -100, -100, 50])
    regions["B"].append([-100, 0, -180, -100])
    regions["B"].append([-100, 0, 50, 180])
    regions["A"].append([-100, 0, -100, 50])
    regions["E"].append([0, 150, -180, -50])
    regions["E"].append([0, 150, 100, 180])
    regions["B"].append([150, 180, 100, 180])
    regions["B"].append([150, 180, -180, -50])
    regions["L"].append([0, 180, -50, 100])

    for region in regions:

        for rama in regions[region]:

            if (rama[0] < phi <= rama[1]) and (
                rama[2] < psi <= rama[3] and math.cos(omega) <= 0
            ):

                rama_type = region

            if (rama[0] < phi <= rama[1]) and (
                rama[2] < psi <= rama[3] and math.cos(omega) > 0
            ):

                region = region.lower()
                rama_type = region

    return rama_type


def get_rot_type(rot):

    rot = float(rot)

    if rot == 999.00:
        return "-"
    elif 0 <= rot < 120:
        return "p"
    elif 120 <= rot < 240:
        return "t"
    elif 240 <= rot < 360:
        return "m"


def calc_dih_angle(
    structure,
    chainid,
    resid_1,
    resid_2,
    resid_3,
    resid_4,
    atomid_1,
    atomid_2,
    atomid_3,
    atomid_4,
    modelid=None,
    check_dist=True,
    check_resid=True,
    check_atomid=True,
    max_dist=2,
):

    if modelid is None:
        modelid = 0

    index_lst = lst_nums(0, 3)

    resid_lst = [resid_1, resid_2, resid_3, resid_4]
    atomid_lst = [atomid_1, atomid_2, atomid_3, atomid_4]

    for i, resid in enumerate(resid_lst):
        resid_lst[i] = resid_to_tuple(resid)

    structure_angle = 999.00

    calc_vect = True

    for index in index_lst:

        resid = resid_lst[index]
        atomid = atomid_lst[index]

        resid_status = True
        if check_resid:
            resid_status = has_resid(structure, chainid, resid, modelid=modelid)

        atomid_status = True
        if resid_status:
            if check_atomid:
                atomid_status = has_atomid(
                    structure, chainid, resid, atomid, modelid=modelid
                )

        if not resid_status or not atomid_status:
            calc_vect = False
            break

    if calc_vect:

        vect_lst = list()

        for index in index_lst:

            resid = resid_lst[index]
            atomid = atomid_lst[index]

            vect = get_atom_vect(
                structure=structure,
                modelid=modelid,
                chainid=chainid,
                resid=resid,
                atomid=atomid,
            )

            vect_lst.append(vect)

        calc_angle = True

        if check_dist:
            for index in index_lst:
                if index != 0 and index != 3:
                    atom_dist = calc_norm_dist(vect_lst[index], vect_lst[index + 1])
                    if atom_dist > max_dist:
                        calc_angle = False

        if calc_angle:
            structure_angle = round(
                math.degrees(
                    calc_dihedral(vect_lst[0], vect_lst[1], vect_lst[2], vect_lst[3])
                ),
                2,
            )

    return structure_angle


def calc_bb_angle(
    structure,
    chainid,
    curr_resid,
    angle,
    modelid=None,
    prev_resid=None,
    next_resid=None,
    check_dist=True,
    check_resid=True,
    check_atomid=True,
    max_dist=2,
):

    atomid_list = bb_dict[angle][0]
    add_list = bb_dict[angle][1]

    index_lst = lst_nums(0, 3)

    if type(curr_resid) == tuple:
        curr_resnum = curr_resid[1]
    else:
        curr_resnum = extract_int(curr_resid)
        curr_resid = resid_to_tuple(curr_resid)

    resid_lst = list()

    check_dist = False

    for index in index_lst:

        add = add_list[index]

        resid = curr_resid

        if add == -1:
            if prev_resid is not None:
                resid = prev_resid
                if type(prev_resid) == tuple:
                    prev_resnum = prev_resid[1]
                if curr_resnum - prev_resnum != 1:
                    check_dist = True
            else:
                resid = resid_to_tuple(curr_resid + add)
        elif add == 1:
            if next_resid is not None:
                resid = next_resid
                if type(next_resid) == tuple:
                    next_resnum = next_resid[1]
                if next_resnum - curr_resnum != 1:
                    check_dist = True
            else:
                resid = resid_to_tuple(curr_resid + add)

        resid_lst.append(resid)

    bb_angle = calc_dih_angle(
        structure=structure,
        modelid=modelid,
        chainid=chainid,
        resid_1=resid_lst[0],
        resid_2=resid_lst[1],
        resid_3=resid_lst[2],
        resid_4=resid_lst[3],
        atomid_1=atomid_list[0],
        atomid_2=atomid_list[1],
        atomid_3=atomid_list[2],
        atomid_4=atomid_list[3],
        check_dist=check_dist,
        check_resid=check_resid,
        check_atomid=check_atomid,
        max_dist=max_dist,
    )

    return bb_angle


def calc_sc_angle(
    structure,
    chainid,
    resid,
    angle,
    modelid=None,
    check_dist=True,
    check_resid=True,
    check_atomid=True,
    max_dist=2,
):

    if modelid is None:
        modelid = 0

    sc_angle = 999.00

    resid = resid_to_tuple(resid)

    if has_resid(structure, chainid, resid, modelid=modelid):

        residue = structure[fix_val(modelid, return_int=True)][chainid][resid]

        resname = get_resname(residue, letter=False)

        sc_angle_dict = sc_dict[angle]

        if resname in list(sc_angle_dict.keys()):

            atomid_lst = sc_dict[angle][resname]

            sc_angle = calc_dih_angle(
                structure=structure,
                modelid=modelid,
                chainid=chainid,
                resid_1=resid,
                resid_2=resid,
                resid_3=resid,
                resid_4=resid,
                atomid_1=atomid_lst[0],
                atomid_2=atomid_lst[1],
                atomid_3=atomid_lst[2],
                atomid_4=atomid_lst[3],
                check_dist=check_dist,
                check_resid=check_resid,
                check_atomid=check_atomid,
                max_dist=max_dist,
            )

            if sc_angle != 999.00:
                if (angle == chi2_col and resname in ["TYR", "PHE", "ASP"]) or (
                    angle == chi3_col and resname == "GLU"
                ):
                    if sc_angle < 0.00:
                        sc_angle += 180.00
                    elif sc_angle > 180.00:
                        sc_angle -= 180.00
                else:
                    if sc_angle < 0.00:
                        sc_angle += 360.00

    return sc_angle



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pandas as pd
from tqdm import tqdm
import concurrent.futures

from ..functions import *


def build_mut_df(df, index, uniprot_dict, resid_lst=None, coord_path_col=None):

    if coord_path_col is None:
        coord_path_col = core_path_col

    df = get_df_at_index(df, index)

    coord_path = df.at[index, coord_path_col]
    modelid = df.at[index, modelid_col]
    chainid = df.at[index, chainid_col]

    structure = load_coord(coord_path)

    uniprot_acc_lst = list(uniprot_dict.keys())

    mut_status_dict = dict()
    mut_pos_dict = dict()
    uniprot_id_lst = list()

    for uniprot_acc in uniprot_acc_lst:

        mut_status_dict[uniprot_acc] = list()
        mut_pos_dict[uniprot_acc] = list()

        uniprot_seq = uniprot_dict[uniprot_acc][seq_col]

        coord_seq = ""
        for resid in lst_nums(1, len(uniprot_seq)):
            resname = "-"
            if has_resid(structure, chainid, resid, modelid=modelid):
                if resname != "X":
                    resname = get_resname(
                        structure[fix_val(modelid, return_int=True)][chainid][resid],
                        letter=True,
                    )
                    ref = uniprot_dict[uniprot_acc][resid]

                    if resname != ref:
                        if resid is not None:
                            add_mut = True
                            if resid_lst is not None:
                                if resid not in resid_lst:
                                    add_mut = False
                            if add_mut:
                                mut_pos = f"{ref}{resid}"
                                mut_status_dict[uniprot_acc].append(mut_pos + resname)
                                mut_pos_dict[uniprot_acc].append(mut_pos)

            coord_seq += resname

        uniprot_id_lst.append(calc_seq_id(coord_seq, uniprot_seq, aln=False))

    uniprot_acc = uniprot_acc_lst[uniprot_id_lst.index(max(uniprot_id_lst))]

    mut_status_lst = mut_status_dict[uniprot_acc]
    mut_pos_lst = mut_pos_dict[uniprot_acc]

    df.at[index, mut_status_col] = lst_to_str(mut_status_lst, empty="WT")
    df.at[index, mut_pos_col] = lst_to_str(mut_pos_lst, empty="WT")
    df.at[index, uniprot_id_col] = uniprot_acc

    return df


def annot_mut(
    df,
    uniprot_accs,
    mut_table_path=None,
    resids=None,
    seq_dir=None,
    coord_path_col=None,
    num_cpu=1,
):

    if type(uniprot_accs) == list:
        uniprot_acc_lst = uniprot_accs
    else:
        uniprot_acc_lst = str_to_lst(uniprot_accs, sep_txt=" ")

    if resids is not None:
        resid_lst = res_to_lst(resids)
    else:
        resid_lst = None

    uniprot_dict = dict()

    for uniprot_acc in uniprot_acc_lst:

        uniprot_dict[uniprot_acc] = dict()

        fasta_url = f"{uniprot_url}{uniprot_acc}.fasta"

        fasta_file = get_seq_path(uniprot_acc, dir_path=seq_dir)

        download_file(fasta_url, fasta_file)

        for record in load_record_lst(fasta_file):

            seq = get_record_seq(record)

        uniprot_dict[uniprot_acc][seq_col] = seq

        for i, resname in enumerate(seq):

            uniprot_dict[uniprot_acc][i + 1] = resname

    mut_df = pd.DataFrame()

    if num_cpu == 1:
        for index in tqdm(
            list(df.index.values), desc="Annotating mutations", position=0, leave=True
        ):
            mut_df = pd.concat(
                [
                    mut_df,
                    build_mut_df(
                        df,
                        index,
                        uniprot_dict,
                        resid_lst=resid_lst,
                        coord_path_col=coord_path_col,
                    ),
                ],
                sort=False,
            )
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    build_mut_df,
                    df,
                    index,
                    uniprot_dict,
                    resid_lst=resid_lst,
                )
                for index in list(df.index.values)
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Annotating mutations",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                mut_df = pd.concat([mut_df, job.result()], sort=False)

    mut_df = mut_df.reset_index(drop=True)

    print("Annotated mutations!")

    if mut_table_path is not None:
        save_table(mut_table_path, mut_df)
    else:
        return mut_df



from src.download.modules import *


def compress_output_files(full_path_to_the_file, gzip_mode="on"):
    if gzip_mode == "on":
        with open(Path(full_path_to_the_file), 'rb') as f_in:
            with gzip.open(Path(str(full_path_to_the_file) + '.gz'), 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from ..functions import *

polymer_color = "gray80"

bio_color = "gray90"
ion_color = "chartreuse"
pharm_color = "salmon"
chem_color = "cyan"
prot_color = "slate"
pocket_color = "grey50"
hb_color = "tv_yellow"
wmhb_color = "tv_blue"

symbol_lst = [" ", "/", "_"]

pymol_lig_col_lst = lig_col_lst + [pocket_lig_col]


def init_pymol(pymol_file, color_dict=None):

    pymol_file.write("bg_color white\n")
    pymol_file.write("set antialias, 2\n")
    pymol_file.write("set orthoscopic, off\n")
    pymol_file.write("set depth_cue, 0\n")
    pymol_file.write("set sphere_scale, 0.4\n")
    pymol_file.write("set stick_radius, 0.2\n")
    pymol_file.write("set hash_max, 300\n")

    if color_dict is not None:
        for group, color in color_dict.items():
            pymol_file.write(f"set_color {group}_color, {color}\n")


def load_obj(pymol_file, path, obj, chainids=None):

    chainid_lst = type_lst(chainids)

    load_cmd = f"load {path}, {obj}"

    if chainid_lst is not None:
        chainids = lst_to_str(chainid_lst, join_txt="+")
        load_cmd += f"; remove model {obj} and not {chainids}//"

    pymol_file.write(f"{load_cmd}\n")


def init_obj(pymol_file, obj=None, ribbon=True, color_chainbow=False):

    color = polymer_color

    if ribbon:
        show = "ribbon"
        hide = "cartoon"
    else:
        show = "cartoon"
        hide = "ribbon"

    pymol_file.write("hide wire, solvent\n")

    if obj is None:
        pymol_file.write(f"show {show}\n")
        pymol_file.write(f"hide {hide}\n")
        pymol_file.write("hide lines\n")
        pymol_file.write("hide sticks\n")
        pymol_file.write("hide spheres\n")
        pymol_file.write(f"color {color}\n")
    else:
        pymol_file.write(f"show {show}, model {obj}\n")
        pymol_file.write(f"hide {hide}, model {obj}\n")
        pymol_file.write(f"hide lines, model {obj}\n")
        pymol_file.write(f"hide sticks, model {obj}\n")
        pymol_file.write(f"hide spheres, model {obj}\n")
        pymol_file.write(f"color {color}, model {obj}\n")

    if color_chainbow:
        pymol_file.write("util.chainbow\n")


def set_color(pymol_file, show_color, color, color_str, return_color=True):

    if show_color == True:
        color_str = color
    elif show_color is not None and show_color != False:
        color_rgb = get_rgb(show_color)
        pymol_file.write(f"set_color {color_str}, {color_rgb}\n")

    if return_color:
        return color_str


def color_sticks(pymol_file, stick_sele, stick_color=None, color_atomic=True):

    if stick_color is not None:
        pymol_file.write(f"color {stick_color}_color, {stick_sele}\n")
    if color_atomic:
        pymol_file.write(f"color atomic, {stick_sele} and (not elem C)\n")


def show_lig(
    pymol_file,
    lig_col,
    lig_dict,
    show_lig,
    lig_color,
    group=None,
    spheres=False,
):

    if show_lig is not None and show_lig != False:

        if group is None:
            lig_str = lst_to_str(lig_dict[lig_col], join_txt="+")
        else:
            lig_str = lst_to_str(lig_dict[group][lig_col], join_txt="+")

        if lig_str is not None:
            if group is None:
                lig_sele = lig_col
                pymol_file.write(f"select {lig_sele}, resn {lig_str}\n")
            else:
                lig_sele = f"{group}_{lig_col}"
                pymol_file.write(f"select {lig_sele}, *_{group} and resn {lig_str}\n")

            pymol_file.write(f"color {lig_color}, {lig_sele}\n")

            if spheres:
                pymol_file.write(f"show spheres, {lig_sele}\n")
            else:
                pymol_file.write(f"show sticks, {lig_sele}\n")
                color_sticks(pymol_file, lig_sele)


def get_sup_sele(obj, chainid, sup_resids=None):

    sup_sele_lst = [f"model {obj} and chain {chainid}"]

    if sup_resids is not None:
        sup_resids = str(sup_resids)
        sele_resid_lst = sup_resids.replace(":", "+")
        sup_sele_lst.append(f"resi {lst_to_str(sele_resid_lst,join_txt='+')}")

    sup_sele_lst.append("bb.")
    sup_sele = lst_to_str(sup_sele_lst, join_txt=" and ")

    return sup_sele


def get_hb_sele(
    x_hb_resname, y_hb_resname, obj=None, x_hb_atomid_str=None, y_hb_atomid_str=None
):
    sele_lst = list()

    if obj is not None:
        sele_lst.append(obj)

    sele_lst.append(x_hb_resname)

    if x_hb_atomid_str is not None:
        sele_lst.append(x_hb_atomid_str.replace("+", "_"))

    sele_lst.append(y_hb_resname)

    if y_hb_atomid_str is not None:
        sele_lst.append(y_hb_atomid_str.replace("+", "_"))

    hb_sele_lst = sele_lst.copy()
    wmhb_sele_lst = sele_lst.copy()
    angle_sele_lst = sele_lst.copy()

    hb_sele_lst.append("hb")
    wmhb_sele_lst.append("wmhb")
    angle_sele_lst.append("angle")

    hb_sele = lst_to_str(hb_sele_lst, join_txt="_")
    wmhb_sele = lst_to_str(wmhb_sele_lst, join_txt="_")
    angle_sele = lst_to_str(angle_sele_lst, join_txt="_")

    return hb_sele, wmhb_sele, angle_sele


def write_pymol_script(
    pymol_df,
    pymol_pml_path,
    group_col=None,
    stick_resids=None,
    loop_resids=None,
    sup_resids=None,
    sup_coord_path=None,
    sup_chainid=None,
    sup_all=True,
    sup_group=False,
    group_order=None,
    style_ribbon=True,
    thick_bb=True,
    color_atomic=True,
    show_resids=None,
    cartoon_transp=0,
    show_bio=None,
    show_ion=None,
    show_pharm=None,
    show_chem=None,
    show_prot=None,
    show_pocket=None,
    color_palette=None,
    color_group=True,
    color_chainbow=False,
    coord_path_col=None,
    prot_chainid_col=None,
    set_view=None,
    eds_dir=None,
    eds_resids=None,
    map_cutoff=1.0,
    diff_cutoff=3.0,
    carve_value=1.6,
    mesh_width=0.5,
    x_hb_resids=None,
    x_hb_atomids=None,
    y_hb_resids=None,
    y_hb_atomids=None,
    max_hb_dist=3.2,
    max_wmhb_dist=3.0,
    show_hb=True,
    show_wmhb=True,
    show_angle=False,
    add_h=False,
):

    df = pymol_df.reset_index(drop=True)

    if coord_path_col is None:
        coord_path_col = core_path_col

    if prot_chainid_col is None:
        prot_chainid_col = bound_prot_chainid_col

    stick_resid_lst = type_lst(stick_resids)
    loop_resid_lst = type_lst(loop_resids)

    x_hb_resid_lst = type_lst(x_hb_resids)
    x_hb_atomid_lst = type_lst(x_hb_atomids)
    y_hb_resid_lst = type_lst(y_hb_resids)
    y_hb_atomid_lst = type_lst(y_hb_atomids)

    if max_wmhb_dist is None:
        max_wmhb_dist = max_hb_dist

    index_lst = list(df.index.values)
    df_col_lst = list(df.columns)

    add_modelid = False
    if len(lst_col(df, modelid_col, unique=True)) > 1:
        add_modelid = True

    label_pdb = False
    if pdb_id_col in df_col_lst:
        label_pdb = True

    if group_col is None:
        group_lst = index_lst
    else:
        for symbol in symbol_lst:
            df[group_col] = df[group_col].str.replace(symbol, "_")
        if group_order is not None:
            for symbol in symbol_lst:
                group_order = [x.replace(symbol, "_") for x in group_order]
            group_lst = group_order
        else:
            group_lst = lst_col(df, group_col, unique=True, return_str=True)

        group_dict = dict()
        for group in group_lst:
            group_dict[group] = list()

    if color_chainbow:
        color_group = False

    if (group_col is None or color_group is False) and loop_resids is not None:
        label_lst = loop_resid_lst.copy()
    else:
        label_lst = group_lst.copy()

    if type(color_palette) == dict:
        color_dict = dict()
        for key, val in color_palette.items():
            for symbol in symbol_lst:
                key = key.replace(symbol, "_")
            color_dict[key] = get_rgb(val)
    else:
        color_dict = get_lst_colors(
            label_lst, palette=color_palette, return_rgb=True, return_dict=True
        )

    append_file_path(pymol_pml_path)

    pymol_file = open(pymol_pml_path, "w")

    init_pymol(pymol_file, color_dict=color_dict)

    bio_color_str = set_color(pymol_file, show_bio, bio_color, "bio_color")
    ion_color_str = set_color(pymol_file, show_ion, ion_color, "ion_color")
    pharm_color_str = set_color(pymol_file, show_pharm, pharm_color, "pharm_color")
    chem_color_str = set_color(pymol_file, show_chem, chem_color, "chem_color")
    prot_color_str = set_color(pymol_file, show_prot, prot_color, "prot_color")
    pocket_color_str = set_color(pymol_file, show_pocket, pocket_color, "pocket_color")

    if type(show_hb) == list:
        hb_color_lst = list()
        for i, color in enumerate(show_hb):
            hb_color_lst.append(set_color(pymol_file, color, hb_color, f"{i}_hb"))

    if type(show_wmhb) == list:
        wmhb_color_lst = list()
        for i, color in enumerate(show_wmhb):
            wmhb_color_lst.append(set_color(pymol_file, color, wmhb_color, f"{i}_wmhb"))

    lig_dict = dict()

    if group_col is None:
        for lig_col in pymol_lig_col_lst:
            lig_dict[lig_col] = list()
    else:
        for group in group_lst:
            lig_dict[group] = dict()
            for lig_col in pymol_lig_col_lst:
                lig_dict[group][lig_col] = list()

    for index in index_lst:

        coord_path = df.at[index, coord_path_col]
        modelid = df.at[index, modelid_col]
        chainid = df.at[index, chainid_col]

        if add_h:
            coord_path = modify_coord_path(coord_path, return_pdb=True, add_h=True)

        obj_lst = list()

        if label_pdb:
            obj_lst.append(df.at[index, pdb_id_col])
            if add_modelid:
                obj_lst.append(modelid)
            if interf_col in df_col_lst:
                obj_lst.append(df.at[index, interf_col])
            if pocket_col in df_col_lst:
                obj_lst.append(df.at[index, pocket_col])
        else:
            obj_lst.append(get_file_name(coord_path))

        group = index
        if group_col is not None:
            group = df.at[index, group_col]
            obj_lst.append(group)

        obj = lst_to_str(obj_lst, join_txt="_")

        if group_col is not None:
            group_dict[str(group)].append([obj, coord_path, chainid])

        for lig_col in pymol_lig_col_lst:
            if lig_col in df_col_lst:
                ligs = df.at[index, lig_col]
                if ligs != "None":
                    for lig in str_to_lst(ligs):
                        if lig != "GLY":
                            if group_col is None:
                                if lig not in lig_dict[lig_col]:
                                    lig_dict[lig_col].append(lig)
                            else:
                                if lig not in lig_dict[group][lig_col]:
                                    lig_dict[group][lig_col].append(lig)

        chainid_lst = type_lst(chainid)
        if show_prot is not None and show_prot != False:
            prot_chainid = df.at[index, prot_chainid_col]
            if prot_chainid != "None":
                prot_chainid_lst = str_to_lst(prot_chainid)
                chainid_lst += prot_chainid_lst

        load_obj(pymol_file, coord_path, obj, chainids=chainid_lst)

        chain_sele = f"model {obj} and chain {chainid}"

        if show_prot is not None and show_prot != False:
            if prot_chainid_col != bound_interf_chainid_col:
                if prot_chainid != "None":
                    prot_chainid_str = lst_to_str(prot_chainid_lst, join_txt="+")

                    prot_sele = f"{obj}_{prot_chainid_str}_{prot_chainid_col}"

                    pymol_file.write(
                        f"select {prot_sele}, model {obj} and chain {prot_chainid_str}\n"
                    )

    init_obj(pymol_file, ribbon=style_ribbon, color_chainbow=color_chainbow)

    if show_resids:
        hide_style = "cartoon"
        if style_ribbon:
            hide_style = "ribbon"
        pymol_file.write(f"hide {hide_style}, not resi {show_resids}\n")

    if stick_resids is not None:
        for stick_resid in stick_resid_lst:
            stick_sele = f"{stick_resid}_stick"
            if group_col is None or color_group is False:
                pymol_file.write(
                    f"select {stick_sele}, resi {stick_resid} and not name O+N+C\n"
                )
                color_sticks(
                    pymol_file,
                    stick_sele,
                    color_atomic=color_atomic,
                )
            else:
                for group in group_lst:
                    group_stick_sele = f"{group}_{stick_sele}"
                    pymol_file.write(
                        f"select {group_stick_sele}, *_{group} and resi {stick_resid} and not name O+N+C\n"
                    )

                    color_sticks(
                        pymol_file,
                        group_stick_sele,
                        stick_color=group,
                        color_atomic=color_atomic,
                    )

                pymol_file.write(f"group {stick_sele}, *_{stick_sele}\n")

            pymol_file.write(f"show sticks, {stick_sele}\n")

    if loop_resids is not None:
        for loop_range in loop_resid_lst:

            loop_range = str(loop_range)

            resid_str = loop_range.replace(":", "+")
            loop_range = loop_range.replace(":", "_")

            loop_sele = f"{loop_range}_loop"

            if group_col is None or color_group is False:
                pymol_file.write(f"select {loop_sele}, resi {resid_str}\n")
                color_sticks(
                    pymol_file,
                    loop_sele,
                    stick_color=loop_range,
                    color_atomic=color_atomic,
                )

                if style_ribbon:
                    pymol_file.write(f"hide cartoon,{loop_sele}\n")
                    if thick_bb:
                        pymol_file.write(f"show sticks, bb. and {loop_sele}\n")
            else:
                for group in group_lst:
                    group_loop_sele = f"{group}_{loop_sele}"
                    pymol_file.write(
                        f"select {group_loop_sele}, *_{group} and resi {resid_str}\n"
                    )
                    color_sticks(
                        pymol_file,
                        group_loop_sele,
                        stick_color=group,
                        color_atomic=color_atomic,
                    )

                    if style_ribbon:
                        pymol_file.write(f"hide cartoon,{group_loop_sele}\n")
                        if thick_bb:
                            pymol_file.write(
                                f"show sticks, bb. and {group_loop_sele}\n"
                            )

                pymol_file.write(f"group {loop_sele}, *_{loop_sele}\n")

    if group_col is None:
        show_lig(
            pymol_file,
            bio_lig_col,
            lig_dict,
            show_bio,
            bio_color_str,
            spheres=False,
        )
        show_lig(
            pymol_file,
            ion_lig_col,
            lig_dict,
            show_ion,
            ion_color_str,
            spheres=True,
        )
        show_lig(
            pymol_file,
            pharm_lig_col,
            lig_dict,
            show_pharm,
            pharm_color_str,
            spheres=False,
        )
        show_lig(
            pymol_file,
            chem_lig_col,
            lig_dict,
            show_chem,
            chem_color_str,
            spheres=False,
        )
        show_lig(
            pymol_file,
            pocket_lig_col,
            lig_dict,
            show_pocket,
            pocket_color_str,
            spheres=True,
        )
    else:
        for group in group_lst:
            show_lig(
                pymol_file,
                bio_lig_col,
                lig_dict,
                show_bio,
                bio_color_str,
                spheres=False,
                group=group,
            )
            show_lig(
                pymol_file,
                ion_lig_col,
                lig_dict,
                show_ion,
                ion_color_str,
                spheres=True,
                group=group,
            )
            show_lig(
                pymol_file,
                pharm_lig_col,
                lig_dict,
                show_pharm,
                pharm_color_str,
                spheres=False,
                group=group,
            )
            show_lig(
                pymol_file,
                chem_lig_col,
                lig_dict,
                show_chem,
                chem_color_str,
                spheres=False,
                group=group,
            )
            show_lig(
                pymol_file,
                pocket_lig_col,
                lig_dict,
                show_pocket,
                pocket_color_str,
                spheres=True,
                group=group,
            )

        if show_bio is not None and show_bio != False:
            pymol_file.write(f"group {bio_lig_col}, *_{bio_lig_col}\n")
        if show_ion is not None and show_ion != False:
            pymol_file.write(f"group {ion_lig_col}, *_{ion_lig_col}\n")
        if show_pharm is not None and show_pharm != False:
            pymol_file.write(f"group {pharm_lig_col}, *_{pharm_lig_col}\n")
        if show_chem is not None and show_chem != False:
            pymol_file.write(f"group {chem_lig_col}, *_{chem_lig_col}\n")
        if show_pocket is not None and show_pocket != False:
            pymol_file.write(f"group {pocket_lig_col}, *_{pocket_lig_col}\n")

    if show_prot is not None and show_prot != False:
        if prot_chainid_col != bound_interf_chainid_col:
            prot_style = "cartoon"
            if style_ribbon:
                prot_style = "ribbon"
            pymol_file.write(f"group {prot_chainid_col}, *_{prot_chainid_col}\n")
            pymol_file.write(f"show {prot_style}, {prot_chainid_col}\n")
            pymol_file.write(f"hide sticks, {prot_chainid_col}\n")
            pymol_file.write(f"color {prot_color_str}, {prot_chainid_col}\n")

    if len(df) > 1:
        if sup_all:
            if sup_coord_path is None:
                sup_coord_path = df.at[0, coord_path_col]
                if sup_chainid is None:
                    sup_chainid = df.at[0, chainid_col]

            if sup_coord_path is not None and sup_chainid is not None:

                sup_obj = "sup_obj"

                load_obj(
                    pymol_file,
                    sup_coord_path,
                    sup_obj,
                    chainids=sup_chainid,
                )

                sup_sele = get_sup_sele(
                    sup_obj,
                    sup_chainid,
                    sup_resids=sup_resids,
                )

                pymol_file.write(f"alignto {sup_sele}\n")
                pymol_file.write(f"center {sup_obj}\n")
                pymol_file.write(f"delete {sup_obj}\n")

    if group_col is not None:
        for group in group_lst:
            if sup_group:
                obj_val_lst = group_dict[group]

                if len(obj_val_lst) > 1:

                    sup_val_lst = obj_val_lst[0]

                    sup_obj = sup_val_lst[0]
                    sup_chainid = sup_val_lst[2]

                    sup_sele = get_sup_sele(
                        sup_obj,
                        sup_chainid,
                        sup_resids=sup_resids,
                    )

                    for val_lst in obj_val_lst:
                        group_obj = val_lst[0]
                        if group_obj != sup_obj:
                            pymol_file.write(f"align {group_obj}, {sup_sele}\n")

            pymol_file.write(f"group {group}, *_{group}\n")

    if x_hb_resid_lst is not None and y_hb_resid_lst is not None:

        pymol_file.write(f"set h_bond_cutoff_edge, {max_hb_dist}\n")

        for index in index_lst:

            coord_path = df.at[index, coord_path_col]
            modelid = df.at[index, modelid_col]
            chainid = df.at[index, chainid_col]

            obj_lst = list()

            if label_pdb:
                obj_lst.append(df.at[index, pdb_id_col])
                if add_modelid:
                    obj_lst.append(modelid)
                if interf_col in df_col_lst:
                    obj_lst.append(df.at[index, interf_col])
                if pocket_col in df_col_lst:
                    obj_lst.append(df.at[index, pocket_col])
            else:
                obj_lst.append(get_file_name(coord_path))

            group = index
            if group_col is not None:
                group = df.at[index, group_col]
                obj_lst.append(group)

            obj = lst_to_str(obj_lst, join_txt="_")

            chain_sele = f"model {obj} and chain {chainid}"

            for i, (x_hb_resname, y_hb_resname) in enumerate(
                zip(x_hb_resid_lst, y_hb_resid_lst)
            ):

                if x_hb_atomid_lst is not None:
                    x_hb_atomid_str = lst_to_str(
                        type_lst(x_hb_atomid_lst[i]), join_txt="+"
                    )
                if y_hb_atomid_lst is not None:
                    y_hb_atomid_str = lst_to_str(
                        type_lst(y_hb_atomid_lst[i]), join_txt="+"
                    )

                hb_sele, wmhb_sele, angle_sele = get_hb_sele(
                    x_hb_resname,
                    y_hb_resname,
                    x_hb_atomid_str=x_hb_atomid_str,
                    y_hb_atomid_str=y_hb_atomid_str,
                    obj=obj,
                )

                x_res_str = "resi"
                y_res_str = "resi"

                if type(x_hb_resname) == str:
                    x_hb_resid = df.at[index, x_hb_resname]
                    x_res_str = "resname"
                else:
                    x_hb_resid = x_hb_resname

                if type(y_hb_resname) == str:
                    y_hb_resid = df.at[index, y_hb_resname]
                    y_res_str = "resname"
                else:
                    y_hb_resid = y_hb_resname

                if x_hb_resid is not None and y_hb_resid is not None:

                    x_hb_resid_sele = f"{chain_sele} and {x_res_str} {x_hb_resid}"
                    y_hb_resid_sele = f"{chain_sele} and {y_res_str} {y_hb_resid}"

                    if x_hb_atomid_lst[i] is not None:
                        x_hb_resid_sele += f" and name {x_hb_atomid_str}"

                    if y_hb_atomid_lst[i] is not None:
                        y_hb_resid_sele += f" and name {y_hb_atomid_str}"

                    write_hb = show_hb
                    if type(show_hb) == list:
                        write_hb = show_hb[i]

                    if write_hb is not False:
                        pymol_file.write(
                            f"dist {hb_sele}, {x_hb_resid_sele}, {y_hb_resid_sele}, cutoff={max_hb_dist}\n"
                        )

                    write_wmhb = show_wmhb
                    if type(show_wmhb) == list:
                        write_wmhb = show_wmhb[i]

                    if write_wmhb is not False:
                        wat_sele_1 = f"byres {chain_sele} and resname HOH within {max_wmhb_dist} of {x_hb_resid_sele}"
                        wat_sele_2 = f"byres {chain_sele} and resname HOH within {max_wmhb_dist} of {y_hb_resid_sele}"

                        wat_sele = f"({wat_sele_1}) in ({wat_sele_2})"

                        pymol_file.write(
                            f"dist {wmhb_sele}, {wat_sele}, ({x_hb_resid_sele})+({y_hb_resid_sele}), cutoff={max_wmhb_dist}\n"
                        )

                        if show_angle:
                            pymol_file.write(
                                f"angle {angle_sele},{x_hb_resid_sele}, {wat_sele}, {y_hb_resid_sele}\n"
                            )

                        pymol_file.write(f"sele HOH, {wat_sele}, merge=1\n")

    if x_hb_resids is not None and y_hb_resids is not None:

        pymol_file.write(f"set sphere_scale, 0.3, HOH\n")

        for i, (x_hb_resname, y_hb_resname) in enumerate(
            zip(x_hb_resid_lst, y_hb_resid_lst)
        ):

            if x_hb_atomid_lst is not None:
                x_hb_atomid_str = lst_to_str(type_lst(x_hb_atomid_lst[i]), join_txt="+")
            if y_hb_atomid_lst is not None:
                y_hb_atomid_str = lst_to_str(type_lst(y_hb_atomid_lst[i]), join_txt="+")

            hb_sele, wmhb_sele, angle_sele = get_hb_sele(
                x_hb_resname,
                y_hb_resname,
                x_hb_atomid_str=x_hb_atomid_str,
                y_hb_atomid_str=y_hb_atomid_str,
            )

            write_hb_status = show_hb
            write_hb_color = hb_color
            if type(show_hb) == list:
                write_hb_status = show_hb[i]
                write_hb_color = hb_color_lst[i]

            write_wmhb_status = show_hb
            write_wmhb_color = wmhb_color
            if type(show_wmhb) == list:
                write_wmhb_status = show_wmhb[i]
                write_wmhb_color = wmhb_color_lst[i]

            if group_col is None or color_group is False:
                if write_hb_status is not False:
                    pymol_file.write(f"group {hb_sele}, *_{hb_sele}\n")
                    pymol_file.write(f"color {write_hb_color}, {hb_sele}\n")
                if write_wmhb_status is not False:
                    pymol_file.write(f"group {wmhb_sele}, *_{wmhb_sele}\n")
                    pymol_file.write(f"color {write_wmhb_color}, {wmhb_sele}\n")
                    if show_angle:
                        pymol_file.write(f"group {angle_sele}, *_{angle_sele}\n")
                        pymol_file.write(f"color {write_wmhb_color}, {angle_sele}\n")
            else:
                for group in group_lst:
                    if write_hb_status is not False:
                        group_hb_sele = f"{group}_{hb_sele}"
                        pymol_file.write(f"group {group_hb_sele}, *_{group_hb_sele}\n")
                        if type(write_hb_status) == str:
                            group_hb_color = write_hb_color
                        else:
                            group_hb_color = f"{group}_color"
                        pymol_file.write(f"color {group_hb_color}, {group_hb_sele}\n")
                    if write_wmhb_status is not False:
                        group_wmhb_sele = f"{group}_{wmhb_sele}"
                        pymol_file.write(
                            f"group {group_wmhb_sele}, *_{group_wmhb_sele}\n"
                        )
                        if type(write_wmhb_status) == str:
                            group_wmhb_color = write_wmhb_color
                        else:
                            group_wmhb_color = f"{group}_color"
                        pymol_file.write(
                            f"color {group_wmhb_color}, {group_wmhb_sele}\n"
                        )
                        if show_angle:
                            group_angle_sele = f"{group}_{angle_sele}"
                            pymol_file.write(
                                f"group {group_angle_sele}, *_{group_angle_sele}\n"
                            )
                            pymol_file.write(
                                f"color {group_wmhb_color}, {group_angle_sele}\n"
                            )

            pymol_file.write(f"set dash_width, 3\n")

            if show_angle:
                pymol_file.write(f"hide angles\n")

            pymol_file.write(f"show spheres, HOH\n")
            pymol_file.write(f"color red, HOH\n")

    if set_view is not None:
        pymol_file.write(f"set_view {set_view}\n")

    if eds_dir is not None and eds_resids is not None:
        if len(df) == 1:
            if pdb_code_col in df_col_lst:
                pdb_code = df.at[0, pdb_code_col]
                chainid = df.at[0, chainid_col]

                map_eds_url = f"{eds_url}{pdb_code}.ccp4"
                diff_eds_url = f"{eds_url}{pdb_code}_diff.ccp4"

                map_path = get_eds_map_path(pdb_code, dir_path=eds_dir)
                diff_path = get_eds_diff_path(pdb_code, dir_path=eds_dir)

                download_file(map_eds_url, map_path)
                download_file(diff_eds_url, diff_path)

                pymol_file.write(f"load {map_path}, {map_str}\n")
                pymol_file.write(f"load {diff_path}, {diff_str}\n")

                mesh_map_sele = f"mesh_{map_str}"
                mesh_diff_sele = f"mesh_{diff_str}"

                eds_resids = str(eds_resids)

                eds_str = eds_resids.replace(":", "+")

                eds_sele = f"{obj} and chain and resi {eds_str}"

                pymol_file.write(
                    f"isomesh {mesh_map_sele}, {map_str}, {map_cutoff}, {eds_sele}, carve={carve_value}\n"
                )
                pymol_file.write(
                    f"isomesh {mesh_diff_sele}, {diff_str}, {diff_cutoff}, {eds_sele}\n"
                )

                pymol_file.write(f"color gray50, {mesh_map_sele}\n")
                pymol_file.write(f"color green, {mesh_diff_sele}\n")
                pymol_file.write("set mesh_negative_color, red\n")
                pymol_file.write("set mesh_negative_visible\n")

                pymol_file.write(f"orient {eds_sele}\n")

                pymol_file.write(f"set mesh_width, {mesh_width}\n")

    if not add_h:
        pymol_file.write("remove hydrogens\n")

    pymol_file.write(f"set cartoon_transparency, {cartoon_transp}\n")

    pymol_file.write("hide labels\n")

    pymol_file.write("zoom center, 25\n")

    pymol_file.close()

    print("Wrote PyMOL script!")



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as np
import matplotlib as mpl
import seaborn as sns

from .table import make_dict
from .lst import move_end_lst

white_hex = "#ffffff"
black_hex = "#000000"

blue_hex = "#1f77b4"
orange_hex = "#ff7f0e"
green_hex = "#2ca02c"
red_hex = "#d62728"
purple_hex = "#9467bd"
brown_hex = "#8c564b"
pink_hex = "#e377c2"
gray_hex = "#7f7f7f"
olive_hex = "#bcbd22"
cyan_hex = "#17becf"

light_blue_hex = "#aec7e8"
light_orange_hex = "#ffbb78"
light_green_hex = "#98df8a"
light_red_hex = "#ff9896"
light_purple_hex = "#c5b0d5"
light_brown_hex = "#c49c94"
light_pink_hex = "#f7b6d2"
light_gray_hex = "#c7c7c7"
light_olive_hex = "#dbdb8d"
light_cyan_hex = "#9edae5"

rama_color_dict = {"A": "#e78ac3", "B": "#8da0cb", "L": "#fc8d62", "E": "#66c2a5"}


def change_hex_alpha(hex_color, alpha):

    return tuple((1.0 - alpha) + np.array(mpl.colors.hex2color(hex_color)) * alpha)


def get_palette_hex_lst(palette, total=None):

    if total is None:
        palette = sns.color_palette(palette)
    else:
        palette = sns.color_palette(palette, total)

    hex_lst = palette.as_hex()

    return hex_lst


def get_hex(color):

    return mpl.colors.to_hex(color)


def get_rgb(color):

    return mpl.colors.to_rgb(color)


def get_hex_lst(color_lst):

    for idx, color in enumerate(color_lst):
        color_lst[idx] = get_hex(color)

    return color_lst


def get_rgb_lst(color_lst):

    for idx, color in enumerate(color_lst):
        color_lst[idx] = get_rgb(color)

    return color_lst


def get_lst_colors(
    label_lst,
    palette=None,
    return_rgb=False,
    return_dict=False,
    alpha=None,
):

    if type(palette) != dict:
        total = len(label_lst)

        if palette is None:

            if total <= 10:

                color_lst = get_palette_hex_lst("tab10")

            elif total <= 20:

                color_lst = get_palette_hex_lst("tab20")

            elif total <= 40:

                color_lst_1 = get_palette_hex_lst("tab20c")
                color_lst_2 = get_palette_hex_lst("tab20b")
                color_lst = color_lst_1 + color_lst_2

            elif total > 40:

                color_lst = get_palette_hex_lst("rainbow", total)

        else:
            if type(palette) == str:
                color_lst = get_palette_hex_lst(palette, total)
            else:
                color_lst = get_hex_lst(palette)

        colors = color_lst[:total]

        if "Noise" in label_lst or "None" in label_lst:
            label_lst = move_end_lst(label_lst, ["Noise", "None"])

            colors[-1] = gray_hex

            if "Noise" in label_lst and "None" in label_lst:
                colors[-2] = gray_hex

        if alpha is not None:
            colors = [change_hex_alpha(x, alpha) for x in colors]

        if return_rgb is True:
            colors = get_rgb_lst(colors)

        if return_dict is True:
            colors = make_dict(label_lst, colors)
    else:
        colors = palette

    return colors



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from Bio import SeqIO, pairwise2
from Bio.Seq import Seq
from Bio.Align import substitution_matrices
from Bio.SeqRecord import SeqRecord
from Bio.Align.Applications import ClustalOmegaCommandline

from .path import delete_path


def load_record_lst(path):

    return SeqIO.parse(path, "fasta")


def load_record_dict(path):

    return SeqIO.to_dict(load_record_lst(path))


def get_record_id(record):

    return str(record.id)


def get_record_seq(record):

    return str(record.seq)


def get_record_desc(record):

    return record.description


def build_record(seq, id, name=None, desc=None):

    return SeqRecord(Seq(seq), id=id, name=name, description=desc)


def clustal_omega(fasta_path, msa_path):

    delete_path(msa_path)

    cline = ClustalOmegaCommandline(infile=fasta_path, outfile=msa_path)
    cline()


def pair_seq_aln(
    ref_seq,
    mob_seq,
    **kwargs,
):

    matrix = kwargs.get("matrix", substitution_matrices.load("BLOSUM62"))
    gap_open = kwargs.get("gap_open", -10.0)
    gap_extend = kwargs.get("gap_extend", -0.5)

    return pairwise2.align.globalds(
        ref_seq, mob_seq, matrix, gap_open, gap_extend, penalize_end_gaps=(False, False)
    )


def calc_seq_id(ref_seq, mob_seq, gap=True, aln=True):

    if aln:
        aln = pair_seq_aln(ref_seq, mob_seq)

        best_aln = aln[0]

        ref_seq = best_aln[0]
        mob_seq = best_aln[1]

    seq_len = min(len(ref_seq), len(mob_seq))
    match_lst = [ref_seq[i] == mob_seq[i] for i in range(seq_len)]

    if gap:
        seq_id = (100 * sum(match_lst)) / seq_len
    else:
        gap_less_seq_len = sum(
            [1 for i in range(seq_len) if (ref_seq[i] != "-" and mob_seq[i] != "-")]
        )
        seq_id = (100 * sum(match_lst)) / gap_less_seq_len

    return seq_id



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from tqdm import tqdm
import concurrent.futures

from ..functions import *


def build_dih_dict(coord_path):

    structure = load_coord(coord_path)

    dih_dict = {coord_path: dict()}

    for model in structure:
        modelid = get_modelid(model)
        dih_dict[coord_path][modelid] = dict()
        for chain in model:
            chainid = get_chainid(chain)

            dih_dict[coord_path][modelid][chainid] = dict()

            resid_lst = [get_resid(residue) for residue in chain if is_aa(residue)]
            resname_lst = [get_resname(residue) for residue in chain if is_aa(residue)]
            resstr_lst = [get_resid_str(residue) for residue in chain if is_aa(residue)]

            for index, resid in enumerate(resid_lst):

                curr_resid = resid
                curr_resname = resname_lst[index]
                curr_resid_str = resstr_lst[index]

                dih_dict[coord_path][modelid][chainid][curr_resid_str] = dict()
                dih_dict[coord_path][modelid][chainid][curr_resid_str][
                    resname_col
                ] = curr_resname

                if index != 0 and index != len(resid_lst) - 1:
                    prev_resid = resid_lst[index - 1]
                    next_resid = resid_lst[index + 1]

                    for bb_col in bb_col_lst:

                        bb_val = calc_bb_angle(
                            structure=structure,
                            modelid=modelid,
                            chainid=chainid,
                            curr_resid=curr_resid,
                            prev_resid=prev_resid,
                            next_resid=next_resid,
                            angle=bb_col,
                        )

                        dih_dict[coord_path][modelid][chainid][curr_resid_str][
                            bb_col
                        ] = bb_val

                else:
                    for bb_col in bb_col_lst:
                        dih_dict[coord_path][modelid][chainid][curr_resid_str][
                            bb_col
                        ] = 999.00

                for sc_col in sc_col_lst:

                    sc_val = calc_sc_angle(
                        structure=structure,
                        modelid=modelid,
                        chainid=chainid,
                        resid=curr_resid,
                        angle=sc_col,
                    )
                    dih_dict[coord_path][modelid][chainid][curr_resid_str][
                        sc_col
                    ] = sc_val

    return dih_dict


def prep_dih(coord_paths, dih_json_path=None, num_cpu=1):

    coord_path_lst = type_lst(coord_paths)

    dih_dict = dict()

    if num_cpu == 1:
        for coord_path in tqdm(
            coord_path_lst,
            desc="Preparing dihedrals",
            position=0,
            leave=True,
        ):
            dih_dict = merge_dicts([dih_dict, build_dih_dict(coord_path)])

    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(build_dih_dict, coord_path)
                for coord_path in coord_path_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Preparing dihedrals",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):
                dih_dict = merge_dicts([dih_dict, job.result()])

    print("Prepared dihedrals!")

    if dih_json_path is not None:
        save_json(dih_json_path, dih_dict)
    else:
        return dih_dict



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from .tools import *
from .functions import *
from .constants import *
from .classify_rascore import *


def update_prep(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    sifts_json_path = get_file_path(sifts_json_file, dir_path=out_path)
    dih_json_path = get_file_path(dih_json_file, dir_path=out_path)

    df = load_table(entry_table_path)

    delete_paths = True
    if past_df is not None:
        if (
            len(
                [
                    x
                    for x in [bound_lig_col, bound_prot_chainid_col]
                    if x not in list(past_df.columns)
                ]
            )
            == 0
        ):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))
            delete_paths = False

    if delete_paths:
        delete_path(sifts_json_path)
        delete_path(dih_json_path)

    if len(df) == 0:
        if past_df is not None:
            save_table(entry_table_path, past_df)
    else:
        pdb_id_lst = lst_col(df, pdb_id_col, unique=True)

        past_sifts_dict = load_json(sifts_json_path)

        prep_coord(
            pdb_id_lst=pdb_id_lst,
            renum_script_path=f"{get_dir_name(__file__)}/PDBrenum/PDBrenum.py",
            coord_table_path=entry_table_path,
            core_dir=out_path,
            rcsb_dir=out_path,
            sifts_dir=out_path,
            renum_dir=out_path,
            sifts_json_path=sifts_json_path,
            data=df,
            lig_resids="3-164",
            update_coords=False,
            num_cpu=num_cpu,
        )

        sifts_dict = load_json(sifts_json_path)
        if past_sifts_dict is not None:
            sifts_dict = merge_dicts([sifts_dict, past_sifts_dict])
        save_json(sifts_json_path, sifts_dict)

        df = load_table(entry_table_path)

        core_path_lst = lst_col(df, core_path_col, unique=True)
        dih_dict = prep_dih(coord_paths=core_path_lst, num_cpu=num_cpu)

        past_dih_dict = load_json(dih_json_path)
        if past_dih_dict is not None:
            dih_dict = merge_dicts([dih_dict, past_dih_dict])
        save_json(dih_json_path, dih_dict)

        if past_df is not None:
            if (
                len(
                    [
                        x
                        for x in [bound_lig_col, bound_prot_chainid_col]
                        if x not in list(past_df.columns)
                    ]
                )
                == 0
            ):
                df = pd.concat([df, past_df], sort=False)
                save_table(entry_table_path, df)


def update_annot(pdbaa_fasta_path, out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)

    df = load_table(entry_table_path)

    if past_df is not None:
        if (
            len(
                [
                    x
                    for x in [mut_class_col, pharm_class_col, prot_class_col, cf_col]
                    if x not in list(past_df.columns)
                ]
            )
            == 0
        ):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))

    if len(df) == 0:
        if past_df is not None:
            save_table(entry_table_path, past_df)
    else:
        df = annot_mut(
            df=df,
            uniprot_accs=uniprot_acc_lst,
            resids="1-166",
            seq_dir=out_path,
            num_cpu=num_cpu,
        )

        for index in list(df.index.values):
            df.at[index, mut_class_col] = lst_to_str(
                [
                    x
                    for x in mut_class_lst
                    if x in str_to_lst(df.at[index, mut_status_col])
                ],
                empty=other_mut_name,
            )

        df = annot_lig(
            df=df,
            lig_dir=out_path,
            site_dict=pharm_site_dict,
            match_dict=pharm_match_dict,
            num_cpu=num_cpu,
        )

        for index in list(df.index.values):
            pharm_class = df.at[index, pharm_lig_site_col]
            if pharm_class not in [sp2_name, sp12_name, none_pharm_name]:
                pharm_class = other_pharm_name
            df.at[index, pharm_class_col] = pharm_class

        df[gene_class_col] = df[prot_col].map(gene_class_dict)
        df[nuc_class_col] = df[bio_lig_col].map(nuc_class_dict).fillna(gtp_name)

        df = annot_prot(
            df=df,
            pdbaa_fasta_path=pdbaa_fasta_path,
            pfam_dict=prot_pfam_dict,
        )

        for index in list(df.index.values):
            prot_class = df.at[index, bound_prot_pfam_col]
            for prot_name, pfam_lst in prot_class_dict.items():
                if prot_class in pfam_lst:
                    prot_class = prot_name
            if prot_class == gef_name:
                nuc_class = df.at[index, nuc_class_col]
                prot_class += "."
                if nuc_class == gtp_name:
                    prot_class += gef_rem_name
                else:
                    prot_class += gef_cdc_name
            df.at[index, prot_class_col] = prot_class

        if (
            past_df is not None
            and len([x for x in class_col_lst if x not in list(past_df.columns)]) == 0
        ):
            df[complete_col] = str(False)
            df = pd.concat([df, past_df], sort=False)
            save_table(entry_table_path, df)

        rcsb_path_lst = lst_col(df, rcsb_path_col, unique=True)
        df = annot_cf(
            coord_paths=rcsb_path_lst,
            data=df,
            num_cpu=num_cpu,
        )

        save_table(entry_table_path, df)


def update_interf(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    interf_json_path = get_file_path(interf_json_file, dir_path=out_path)
    interf_table_path = get_file_path(interf_table_file, dir_path=out_path)

    df = load_table(entry_table_path)

    delete_paths = True
    if past_df is not None and interf_class_col in list(past_df.columns):
        df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))
        delete_paths = False

    if delete_paths:
        delete_path(interf_json_path)
        delete_path(interf_table_path)

    if len(df) == 0:
        if past_df is not None:
            save_table(entry_table_path, past_df)
    else:
        xray_df = mask_equal(df, method_col, "XRAY")

        coord_path_lst = lst_col(xray_df, renum_path_col, unique=True)

        chainid_dict = dict()
        for coord_path in coord_path_lst:
            chainid_dict[coord_path] = list()
        for index in list(xray_df.index.values):
            coord_path = xray_df.at[index, renum_path_col]
            chainid = xray_df.at[index, chainid_col]
            chainid_dict[coord_path].append(chainid)

        interf_dict = prep_interf(
            coord_paths=coord_path_lst,
            interf_dir=out_path,
            chainid_dict=chainid_dict,
            num_cpu=num_cpu,
        )

        past_interf_dict = load_json(interf_json_path)
        if past_interf_dict is not None:
            interf_dict = merge_dicts([interf_dict, past_interf_dict])
        save_json(interf_json_path, interf_dict)

        interf_df = build_interf_table(
            df=xray_df,
            interf_dict=interf_dict,
            search_coord_path=get_renum_path(sup_pdb_code, dir_path=out_path),
            search_chainid=sup_chainid,
            search_interf="4",
        )

        past_interf_df = load_table(interf_table_path)
        if past_interf_df is not None and interf_class_col in list(past_df.columns):
            interf_df = pd.concat([interf_df, past_interf_df], sort=False)
        save_table(interf_table_path, interf_df)

        interf_pdb_id_lst = lst_col(interf_df, pdb_id_col)
        df[interf_class_col] = (
            df[pdb_id_col]
            .map(make_dict(interf_pdb_id_lst, len(interf_pdb_id_lst) * [dimer_name]))
            .fillna(none_dimer_name)
        )

        if past_df is not None:
            if interf_class_col in list(past_df.columns):
                df[complete_col] = str(False)
                df = pd.concat([df, past_df], sort=False)

        save_table(entry_table_path, df)


def update_pocket(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    pocket_json_path = get_file_path(pocket_json_file, dir_path=out_path)
    pocket_table_path = get_file_path(pocket_table_file, dir_path=out_path)

    df = load_table(entry_table_path)

    delete_paths = True
    if past_df is not None:
        if pocket_class_col in list(past_df.columns):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))
            delete_paths = False

    if delete_paths:
        delete_path(pocket_json_path)
        delete_path(pocket_table_path)

    if len(df) == 0:
        if past_df is not None:
            save_table(entry_table_path, past_df)
    else:
        coord_path_lst = [x.replace(".cif", ".pdb") for x in lst_col(df, core_path_col)]

        pharm_dict = dict()
        chainid_dict = dict()
        for coord_path in coord_path_lst:
            chainid_dict[coord_path] = list()
        for index in list(df.index.values):
            coord_path = df.at[index, core_path_col].replace(".cif", ".pdb")
            chainid = df.at[index, chainid_col]
            chainid_dict[coord_path].append(chainid)
            if df.at[index, pharm_lig_site_col] in [sp2_name, sp12_name]:
                pharm_dict[coord_path] = df.at[index, pharm_lig_col]

        pocket_dict = prep_pocket(
            coord_paths=coord_path_lst,
            pocket_dir=out_path,
            pharm_dict=pharm_dict,
            chainid_dict=chainid_dict,
            num_cpu=num_cpu,
        )

        past_pocket_dict = load_json(pocket_json_path)
        if past_pocket_dict is not None:
            pocket_dict = merge_dicts([pocket_dict, past_pocket_dict])
        save_json(pocket_json_path, pocket_dict)

        if past_df is not None:
            if pocket_class_col in list(past_df.columns):
                df[complete_col] = str(False)
                df = pd.concat([df, past_df], sort=False)

        temp_df = build_pocket_table(
            df=df,
            pocket_dict=pocket_dict,
        )
        temp_df = mask_equal(
            temp_df, pocket_type_col, [title_str(pharm_lig_col), pocket_unbound_name]
        )

        pocket_df = pd.DataFrame()
        for pharm_name in [sp2_name, sp12_name]:
            bound_df = mask_equal(temp_df, pharm_class_col, pharm_name)
            bound_df = mask_equal(bound_df, pocket_status_col, pocket_bound_name)

            search_cont_lst = lst_col(
                mask_equal(bound_df, pharm_class_col, pharm_name), pharm_lig_cont_col
            )
            search_cont_lst = [str_to_lst(x) for x in search_cont_lst]

            unbound_df = build_pocket_table(
                df=mask_equal(df, pharm_class_col, none_pharm_name),
                pocket_dict=pocket_dict,
                search_cont_lst=search_cont_lst,
                search_max_dist=0.4,
                use_simpson=True,
            )

            unbound_df = mask_equal(unbound_df, pocket_status_col, pocket_unbound_name)

            bound_df[pocket_site_col] = pharm_name
            unbound_df[pocket_site_col] = pharm_name

            pocket_df = pd.concat([pocket_df, bound_df], sort=False)
            pocket_df = pd.concat([pocket_df, unbound_df], sort=False)

        other_df = mask_unequal(
            temp_df, pocket_id_col, lst_col(pocket_df, pocket_id_col)
        )
        other_df = mask_equal(other_df, pocket_status_col, pocket_unbound_name)
        other_df = mask_equal(other_df, pharm_class_col, none_pharm_name)

        other_df[pocket_site_col] = other_pharm_name

        pocket_df = pd.concat([pocket_df, other_df], sort=False)

        save_table(pocket_table_path, pocket_df)

        pocket_df = load_table(pocket_table_path)
        annot_df = mask_equal(pocket_df, pocket_site_col, [sp2_name, sp12_name])

        for index in tqdm(
            list(df.index.values),
            desc="Annotating pockets",
            position=0,
            leave=True,
        ):
            pdb_df = mask_equal(annot_df, pdb_id_col, df.at[index, pdb_id_col])

            pocket_site_lst = lst_col(pdb_df, pocket_site_col, unique=True)
            pocket_status_lst = lst_col(pdb_df, pocket_status_col, unique=True)

            if len(pocket_site_lst) > 1:
                if other_pharm_name in pocket_site_lst:
                    pocket_site_lst.remove(other_pharm_name)

            if len(pocket_site_lst) == 0:
                pharm_class = df.at[index, pharm_class_col]
                if pharm_class in [sp2_name, sp12_name, other_pharm_name]:
                    pocket_site_lst = [pharm_class]
                    pocket_status_lst = [pocket_bound_name]
                else:
                    pocket_site_lst = [other_pharm_name]
                    pocket_status_lst = [pocket_unbound_name]

            pocket_site = lst_to_str(pocket_site_lst, join_txt="|")
            pocket_status = lst_to_str(pocket_status_lst)

            df.at[index, pocket_site_col] = pocket_site
            df.at[index, pocket_status_col] = pocket_status

            pocket_class = pocket_site

            pocket_class += "-"
            pocket_class += pocket_status

            df.at[index, pocket_class_col] = pocket_class

        save_table(entry_table_path, df)


def update_classify(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    dih_json_path = get_file_path(dih_json_file, dir_path=out_path)

    df = load_table(entry_table_path)

    if past_df is not None:
        if sw1_name in list(past_df.columns) and sw2_name in list(past_df.columns):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))

    dih_dict = load_json(dih_json_path)

    classify_path = f"{out_path}/{classify_str}"

    if len(df) > 0:
        classify_rascore(
            df,
            out_path=classify_path,
            dih_dict=dih_dict,
            num_cpu=num_cpu,
        )

    result_df = load_table(get_file_path(result_table_file, dir_path=classify_path))

    for loop_name in list(loop_resid_dict.keys()):
        df[loop_name] = df[pdb_id_col].map(
            make_dict(lst_col(result_df, pdb_id_col), lst_col(result_df, loop_name))
        )

    if past_df is not None:
        if sw1_name in list(past_df.columns) and sw2_name in list(past_df.columns):
            df[complete_col] = str(False)
            df = pd.concat([df, past_df], sort=False)

    save_table(entry_table_path, df)


def build_rascore(out_path=None, pdbaa_fasta_path=None, num_cpu=1):

    if out_path is None:
        out_path = f"{os.getcwd()}/{rascore_str}_{build_str}"

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)

    if pdbaa_fasta_path is None:
        curr_date = datetime.today().strftime("%Y-%m-%d")
        pdbaa_fasta_path = get_file_path(
            f"{curr_date}_{pdbaa_fasta_file}",
            dir_str=pdbaa_str,
            dir_path=out_path,
            pre_str=False,
        )

    past_df = load_table(entry_table_path)

    if past_df is None:
        print("No rascore database found! Building rascore database from scratch!")
    else:
        if complete_col in list(past_df.columns):
            past_df = mask_equal(past_df, complete_col, str(True))
        else:
            past_df[complete_col] = str(True)

    print("Downloading updated pdbaa file.")

    try:
        download_unzip(url=pdbaa_url, path=pdbaa_fasta_path)
    except:
        pdbaa_fasta_path_lst = [
            x
            for x in os.listdir(get_dir_path(dir_str=pdbaa_str, dir_path=out_path))
            if pdbaa_fasta_file in x
        ]
        if len(pdbaa_fasta_path_lst) > 0:
            pdbaa_fasta_path = sorted(
                pdbaa_fasta_path_lst,
                reverse=True,
            )[0]

            last_date = pdbaa_fasta_path.split(f"_{pdbaa_fasta_file}")[0]
            last_date = last_date.split(f"{pdbaa_str}/")[1]
            print(f"Updated pdbaa file unavailable. Using latest version ({last_date})")

    try:
        search_pdbaa(
            pdbaa_fasta_path=pdbaa_fasta_path,
            search_lst=swiss_id_lst,
            entry_table_path=entry_table_path,
            min_length=25,
        )
    except:
        print("Error reading pdbaa file.")
        copy_path(
            get_file_path(
                entry_table_file,
                get_dir_path(dir_str=data_str, dir_path=get_dir_name(__file__)),
            ),
            entry_table_path,
        )

    if past_df is not None:
        df = load_table(entry_table_path)
        if len(df) < len(past_df):
            print("Downgrading rascore database to older version.")
            past_df = None

    update_prep(out_path=out_path, past_df=past_df, num_cpu=num_cpu)
    update_annot(
        pdbaa_fasta_path,
        out_path=out_path,
        past_df=past_df,
        num_cpu=num_cpu,
    )
    update_interf(out_path=out_path, past_df=past_df, num_cpu=num_cpu)
    update_pocket(out_path=out_path, past_df=past_df, num_cpu=num_cpu)
    update_classify(out_path=out_path, past_df=past_df, num_cpu=num_cpu)

    df = load_table(entry_table_path)
    if complete_col in list(df.columns):
        del df[complete_col]

    save_table(entry_table_path, df)

    copy_path(
        entry_table_path,
        get_file_path(
            entry_table_file,
            dir_path=get_dir_path(
                dir_str=data_str,
                dir_path=get_dir_name(__file__),
            ),
        ),
    )

    print("Rascore update complete!")


# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2022 Mitchell Isaac Parker

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from tqdm import tqdm

from ..functions import *


def annot_prot(
    df,
    pdbaa_fasta_path,
    pfam_dict=None,
    site_dict=None,
    other_pfam="Other",
    other_site="Unclassified",
    none_pfam="None",
    none_site="None",
    max_site_dist=5,
    prot_table_path=None,
):

    pdbaa_dict = load_record_dict(pdbaa_fasta_path)

    if site_dict is not None:
        for site_name, site_cont in site_dict.items():
            site_dict[site_name] = res_to_lst(site_cont)

    if pfam_dict is not None:
        pfam_dict_lst = list(pfam_dict.keys())

    for index in tqdm(
        list(df.index.values), desc="Annotating proteins", position=0, leave=True
    ):

        pdb_code = df.at[index, pdb_code_col]

        bound_prot_chainid = df.at[index, bound_prot_chainid_col]

        prot_lst = list()
        swiss_id_lst = list()

        cont_lst = list()
        site_lst = list()
        pfam_lst = list()

        if bound_prot_chainid != "None":
            coord_path = df.at[index, core_path_col]
            modelid = df.at[index, modelid_col]
            chainid = df.at[index, chainid_col]
            structure = load_coord(coord_path)
            neighbors = get_neighbors(
                structure[fix_val(modelid, return_int=True)][chainid]
            )
            for bound_chainid in str_to_lst(bound_prot_chainid):
                record = pdbaa_dict[f"{pdb_code.upper()}{bound_chainid}"]
                desc = get_record_desc(record)
                prot = get_pdbaa_prot(desc)
                swiss_id = get_pdbaa_swiss_id(desc)

                prot_cont_lst = lst_unique(
                    [
                        get_resnum(x)
                        for x in get_chain_cont(
                            neighbors,
                            structure[fix_val(modelid, return_int=True)][bound_chainid],
                            level="R",
                            max_dist=max_site_dist,
                        )
                        if (get_resnum(x) < 50000)
                        and (get_reschainid(x) == chainid)
                        and is_aa(x)
                    ]
                )
                cont_lst += prot_cont_lst

                if prot not in prot_lst:
                    prot_lst.append(prot)
                    swiss_id_lst.append(swiss_id)

                if site_dict is not None:

                    max_cont = 0
                    site_status = other_site
                    for site_name, site_cont_lst in site_dict.items():
                        site_cont = len(
                            [x for x in site_cont_lst if x in prot_cont_lst]
                        )
                        if site_cont > max_cont:
                            site_status = site_name
                            max_cont = site_cont
                    site_lst.append(site_status)

                if pfam_dict is not None:
                    pfam_status = other_pfam
                    if swiss_id in pfam_dict_lst:
                        pfam_status = pfam_dict[swiss_id]
                    elif swiss_id == "NA":
                        pfam_status = "Binder"
                    pfam_lst.append(pfam_status)

        df.at[index, bound_prot_col] = lst_to_str(lst_unique(prot_lst))
        df.at[index, bound_prot_swiss_id_col] = lst_to_str(lst_unique(swiss_id_lst))
        df.at[index, bound_prot_site_col] = lst_to_str(
            lst_unique(site_lst), empty=none_site
        )
        df.at[index, bound_prot_pfam_col] = lst_to_str(
            lst_unique(pfam_lst), empty=none_pfam
        )
        df.at[index, bound_prot_cont_col] = lst_to_str(
            lst_unique(cont_lst), empty=none_pfam
        )

    print("Annotated proteins!")

    if prot_table_path is not None:
        save_table(prot_table_path, df)
    else:
        return df


def long_usage_messenger():
    return '''
PDB.py
optional arguments:
-h, --help            show this help message and exit

-rftf text_file_with_PDB.txt, --renumber_from_text_file text_file_with_PDB.txt
This option will download and renumber specified files
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.txt -mmCIF
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.txt -PDB
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.txt -mmCIF_assembly
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.txt -PDB_assembly
usage $ python3 PDB.py -rftf text_file_with_PDB_in_it.txt -all

-rfla [6dbp 3v03 2jit ...], --renumber_from_list_of_arguments [6dbp 3v03 2jit ...]
This option will download and renumber specified files
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -PDB
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF_assembly
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -PDB_assembly
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -all

-dftf text_file_with_PDB.txt, --download_from_text_file text_file_with_PDB.txt
This option will read given input file parse by space
or tab or comma or new line and download it example 
usage $ python3 PDB.py -dftf text_file_with_PDB_in_it.txt -mmCIF
usage $ python3 PDB.py -dftf text_file_with_PDB_in_it.txt -PDB
usage $ python3 PDB.py -dftf text_file_with_PDB_in_it.txt -mmCIF_assembly
usage $ python3 PDB.py -dftf text_file_with_PDB_in_it.txt -PDB_assembly
usage $ python3 PDB.py -dftf text_file_with_PDB_in_it.txt -all

-dfla [6dbp 3v03 2jit ...], --download_from_list_of_arguments 6dbp 3v03 2jit ...]
This option will read given list of arguments separated by space. 
Format of the list should be without any commas or quotation marks
usage $ python3 PDB.py -dfla 6dbp 3v03 2jit -mmCIF
usage $ python3 PDB.py -dfla 6dbp 3v03 2jit -PDB
usage $ python3 PDB.py -dfla 6dbp 3v03 2jit -mmCIF_assembly
usage $ python3 PDB.py -dfla 6dbp 3v03 2jit -PDB_assembly
usage $ python3 PDB.py -dfla 6dbp 3v03 2jit -all

-redb, --renumber_entire_database
This option will download and renumber entire PDB database in PDB or/and mmCIF format
usage $ python3 PDB.py -redb -mmCIF
usage $ python3 PDB.py -redb -PDB
usage $ python3 PDB.py -redb -mmCIF_assembly
usage $ python3 PDB.py -redb -PDB_assembly
usage $ python3 PDB.py -redb -all 

-dall, --download_entire_database
This option will download entire mmCIF database
usage $ python3 PDB.py -dall -mmCIF
usage $ python3 PDB.py -dall -PDB
usage $ python3 PDB.py -dall -mmCIF_assembly
usage $ python3 PDB.py -dall -PDB_assembly
usage $ python3 PDB.py -dall -all

-refr, --refresh_entire_database
This option will delete outdated files and download
fresh ones. This option makes sense and only works if
you work with entire database
usage $ python3 PDB.py -refr -mmCIF
usage $ python3 PDB.py -refr -PDB
usage $ python3 PDB.py -refr -mmCIF_assembly
usage $ python3 PDB.py -refr -PDB_assembly
usage $ python3 PDB.py -refr -all    

-PDB, --PDB_format_only
This option will specify working format to pdb format
-mmCIF, --mmCIF_format_only
This option will specify working format to mmCIF format (default)
-PDB_assembly, --PDB_assembly_format_only
This option will specify working format to pdb format
-mmCIF_assembly, --mmCIF_assembly_format_only
This option will specify working format to mmCIF format
-all, --all_formats   This option will work with both formats

argpar.add_argument("-sipm", "--set_default_input_path_to_mmCIF", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-sipma", "--set_default_input_path_to_mmCIF_assembly", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-sipp", "--set_default_input_path_to_PDB", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-sippa", "--set_default_input_path_to_PDB_assembly", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-sips", "--set_default_input_path_to_SIFTS", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-sopm", "--set_default_output_path_to_mmCIF", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-sopma", "--set_default_output_path_to_mmCIF_assembly", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-sopp", "--set_default_output_path_to_PDB", type=str, help=argparse.SUPPRESS)
argpar.add_argument("-soppa", "--set_default_output_path_to_PDB_assembly", type=str, help=argparse.SUPPRESS)

-sipm, --set_default_input_path_to_mmCIF
This option will set default input path to mmCIF files (default: <./mmCIF>)
usage $ python3 PDB.py -sipm /Users/bulatfaezov/PycharmProjects/renum/venv/mmCIF
-sipp, --set_default_input_path_to_PDB
This option will set default input path to PDB files (default: <./PDB>)
usage $ python3 PDB.py -sipp /Users/bulatfaezov/PycharmProjects/renum/venv/PDB
-sipma, --set_default_input_path_to_mmCIF_assembly
This option will set default input path to mmCIF_assembly files (default: <./mmCIF_assembly>)
usage $ python3 PDB.py -sipm /Users/bulatfaezov/PycharmProjects/renum/venv/mmCIF_assembly
-sippa, --set_default_input_path_to_PDB_assembly
This option will set default input path to PDB_assembly files (default: <./PDB_assembly>)
usage $ python3 PDB.py -sipp /Users/bulatfaezov/PycharmProjects/renum/venv/PDB_assembly
-sips, --set_default_input_path_to_SIFTS
This option will set default input path to SIFTS files (default: <./SIFTS>)
usage $ python3 PDB.py -sips /Users/bulatfaezov/PycharmProjects/renum/venv/SIFTS
-sopm, --set_default_output_path_to_mmCIF
This option will set default output path to mmCIF files (default: <./output_mmCIF>)
usage $ python3 PDB.py -sopm /Users/bulatfaezov/PycharmProjects/renum/venv/output_mmCIF
-sopp, --set_default_output_path_to_PDB
This option will set default output path to PDB files (default: <./output_PDB>)
usage $ python3 PDB.py -sopp /Users/bulatfaezov/PycharmProjects/renum/venv/output_PDB
-sopma, --set_default_output_path_to_mmCIF_assembly
This option will set default output path to mmCIF_assembly files (default: <./output_mmCIF_assembly>)
usage $ python3 PDB.py -sopm /Users/bulatfaezov/PycharmProjects/renum/venv/output_mmCIF_assembly
-soppa, --set_default_output_path_to_PDB_assembly
This option will set default output path to PDB_assembly files (default: <./output_PDB_assembly>)
usage $ python3 PDB.py -sopp /Users/bulatfaezov/PycharmProjects/renum/venv/output_PDB_assembly

-sdmn, --set_default_mmCIF_num
This option will set default mmCIF number which will be added to 1 to end numbering in cases 
when there are no UniProt numbering (default: 50000)
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF -sdmn 50000

-sdpn, --set_default_PDB_num
This option will set default PDB number which will be added to 1 to end numbering in cases 
when there are no UniProt numbering (default: 5000)
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF -sdpn 5000

"-offz", "--set_to_off_mode_gzip"
By default program will compress files with gzip this option will turn that off
(default: gzip is on)
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF -offz

"-nproc", "--set_number_of_processes"
By default program will use all available CPUs. User can reduce number of CPUs for PDBrenum.
In this example: only 4 CPUs will be used by the PDBrenum even if more CPUs available
(default: nproc = None)
usage $ python3 PDB.py -rfla 6dbp 3v03 2jit -mmCIF -nproc 4


Roland Dunbrack's Lab
Fox Chase Cancer Center
Philadelphia, PA
2020
        '''
